#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import cherrypy
from cocktail.translations import translations
from cocktail.pkgutils import resolve
from cocktail import schema
from cocktail.persistence import transaction
from cocktail.controllers import request_property, context, Location
from woost import app
from woost.models import changeset_context, ModifyPermission
from woost.controllers.backoffice.basebackofficecontroller \
    import BaseBackOfficeController
from woost.controllers.backoffice.useractions import (
    UserAction,
    get_user_action
)
from .request import TranslationWorkflowRequest
from .transitionpermission import TranslationWorkflowTransitionPermission
from .transitionnode import TranslationWorkflowTransitionNode
from .utils import (
    member_is_included_in_translation_workflow,
    notify_translation_request_changes
)


class TranslationWorkflowTransitionAction(UserAction):

    content_type = TranslationWorkflowRequest
    min = 1
    max = None
    included = frozenset([
        "toolbar",
        "item_buttons"
    ])
    show_as_primary_action = "on_content_type"
    hidden_when_disabled = True
    _transition_id = None

    def __translate__(self, language, **kwargs):
        return translations(self.transition, language, **kwargs)

    @request_property
    def transition(self):
        from .transition import TranslationWorkflowTransition
        return TranslationWorkflowTransition.get_instance(self._transition_id)

    @request_property
    def icon_uri(self):
        icon = self.transition.icon
        return (
            icon
            and icon.get_uri()
            or "/resources/images/translation_workflow_transition.png"
        )

    def is_available(self, context, target):

        if isinstance(target, TranslationWorkflowRequest):
            transition = self.transition

            if (
                transition.requires_different_state
                and target.state is transition.target_state
            ):
                return False

            if (
                transition.source_states
                and target.state not in transition.source_states
            ):
                return False

        return UserAction.is_available(self, context, target)

    def is_permitted(self, user, target):
        return user.has_permission(
            TranslationWorkflowTransitionPermission,
            translation_request = None if isinstance(target, type) else target,
            transition = self.transition
        )

    def invoke(self, controller, selection):

        transition_data = None
        user = app.user

        # Save the edit stack
        if (
            len(selection) == 1
            and controller.edit_node
            and isinstance(
                controller.edit_node.item,
                TranslationWorkflowRequest
            )
            and hasattr(controller, "save_item")
            and user.has_permission(
                ModifyPermission,
                target = selection[0]
            )
        ):
            try:
                controller.save_item()
            except cherrypy.HTTPRedirect:
                pass

        # Initialize the transition process. If the transition requires
        # parameters, spawn a new node in the edit stack and redirect the user
        # to a form.
        if self.transition.transition_setup_class:
            transition_setup = (
                resolve(self.transition.transition_setup_class)
                (self.transition, selection)
            )
            if transition_setup.multiple_choices:
                node = TranslationWorkflowTransitionNode()
                node.requests = selection
                node.transition = self.transition
                node.push_to_stack()
                node.go()

            transition_schema = transition_setup.transition_schema
            if transition_schema is not None:
                transition_data = {}
                transition_schema.init_instance(transition_data)

        # Otherwise, execute the transition right away
        def execute_transition():
            transition = self.transition
            with changeset_context(author = user) as changeset:
                for request in selection:
                    transition.execute(request, transition_data)
                    change = changeset.changes.get(request.id)
                    if change is not None:
                        change.is_explicit_change = True
            return changeset

        changeset = transaction(execute_transition)
        notify_translation_request_changes(changeset)
        Location.get_current().go("GET")

    def get_errors(self, controller, selection):

        for error in UserAction.get_errors(self, controller, selection):
            yield error

        # Some transitions (ie. "apply") can't proceed unless the text in the
        # request is complete and valid
        if self.transition.requires_valid_text:
            stack_node = getattr(controller, "stack_node", None)

            # Applying the transition on a request that is being edited:
            # validate the edit state
            if stack_node is not None and selection == [stack_node.item]:
                adapter = schema.Adapter()
                adapter.implicit_copy = False
                for member in stack_node.form_schema.iter_members():
                    if member.name.startswith("translated_values_"):
                        adapter.copy(
                            member.name,
                            properties = {"required": True}
                        )
                val_schema = adapter.export_schema(stack_node.form_schema)
                for error in val_schema.get_errors(
                    stack_node.form_data,
                    languages = stack_node.item_translations,
                    persistent_object = stack_node.item
                ):
                    yield TranslationWorkflowInvalidTextError(
                        stack_node.item,
                        self.transition,
                        error
                    )

            # Applying the transition outside an edit form
            else:
                val_schemas = {}

                for request in selection:
                    translated_model = request.translated_item.__class__
                    val_schema = val_schemas.get(translated_model)

                    if val_schema is None:
                        adapter = schema.Adapter()
                        adapter.implicit_copy = False
                        for member in translated_model.iter_members():
                            if member_is_included_in_translation_workflow(member):
                                adapter.copy(
                                    member.name,
                                    properties = {
                                        "required": True,
                                        "translated": False
                                    }
                                )
                        val_schema = adapter.export_schema(translated_model)
                        val_schemas[translated_model] = val_schema

                    for error in val_schema.get_errors(
                        request.translated_values
                    ):
                        yield TranslationWorkflowInvalidTextError(
                            request,
                            self.transition,
                            error
                        )

    @classmethod
    def register_transition_action(cls, transition):
        action = cls.get_transition_action(transition)
        if action is None:
            action = cls("translation_workflow_transition_%d" % transition.id)
            action._transition_id = transition.id
            action.register()

    @classmethod
    def get_transition_action(cls, transition):
        return get_user_action(
            "translation_workflow_transition_%d" % transition.id
        )


class TranslationWorkflowInvalidTextError(Exception):

    def __init__(self, request, transition, error):
        self.request = request
        self.transition = transition
        self.error = error


BaseBackOfficeController._graceful_user_action_errors.add(
    TranslationWorkflowInvalidTextError
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.html import templates

SlideShowBlockView = templates.get_class("woost.views.SlideShowBlockView")




#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			October 2008
"""
from cocktail.events import event_handler
from cocktail import schema
from cocktail.modeling import getter
from .item import Item


class Style(Item):

    type_group = "customization"

    members_order = [
        "title",
        "applicable_to_text",
        "applicable_to_blocks",
        "custom_class_name",
        "declarations",
        "admin_declarations"
    ]

    title = schema.String(
        required = True,
        unique = True,
        indexed = True,
        normalized_index = True,
        full_text_indexed = True,
        descriptive = True,
        translated = True,
        spellcheck = True
    )

    applicable_to_text = schema.Boolean(
        required = True,
        indexed = True,
        default = True,
        listed_by_default = False
    )

    applicable_to_blocks = schema.Boolean(
        required = True,
        default = True,
        indexed = True,
        listed_by_default = False
    )

    custom_class_name = schema.String(
        indexed = True,
        unique = True,
        normalized_index = False,
        listed_by_default = False
    )

    declarations = schema.String(
        text_search = False,
        edit_control = "cocktail.html.TextArea",
        listed_by_default = False
    )

    admin_declarations = schema.String(
        text_search = False,
        edit_control = "cocktail.html.TextArea",
        listed_by_default = False
    )

    @property
    def class_name(self):
        return self.custom_class_name or ("woost_style_%d" % self.id)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Item, File


class TranslationWorkflowState(Item):

    type_group = "translation_workflow"

    members_order = [
        "title",
        "plural_title",
        "color",
        "incomming_transitions",
        "outgoing_transitions"
    ]

    title = schema.String(
        translated = True,
        required = True,
        unique = True,
        indexed = True,
        normalized_index = True,
        descriptive = True,
        spellcheck = True
    )

    plural_title = schema.String(
        translated = True,
        required = True,
        unique = True,
        indexed = True,
        normalized_index = True,
        listed_by_default = False,
        spellcheck = True
    )

    color = schema.Color(
        required = True,
        default = "#ffffff",
        listed_by_default = False
    )

    incomming_transitions = schema.Collection(
        items = "woost.extensions.translationworkflow."
                "transition.TranslationWorkflowTransition",
        related_key = "target_state",
        bidirectional = True,
        cascade_delete = True
    )

    outgoing_transitions = schema.Collection(
        items = "woost.extensions.translationworkflow."
                "transition.TranslationWorkflowTransition",
        related_key = "source_states",
        bidirectional = True,
        cascade_delete = True
    )

    state_after_source_change = schema.Reference(
        items = schema.Reference(),
        listed_by_default = False
    )


# Self reference for state_after_source_change
TranslationWorkflowState.state_after_source_change.type = \
    TranslationWorkflowState

TranslationWorkflowState.state_after_source_change.related_end = \
    schema.Collection()




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Pepe Osca <pepe.osca@whads.com>
"""
from cocktail import schema
from woost.models import User

User.add_member(
    schema.String("google_user_id",
        indexed = True,
        unique = True,
        editable = schema.READ_ONLY,
        listed_by_default = False
    )
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from warnings import warn
from collections import Sequence, Set, defaultdict, Counter
from datetime import date, time, datetime
from contextlib import contextmanager
import sys
import base64
import json
import argparse
import enum
from cocktail.typemapping import TypeMapping
from cocktail.styled import styled, ProgressBar
from cocktail.pkgutils import get_full_name, resolve
from cocktail.modeling import (
    ListWrapper,
    SetWrapper,
    GenericMethod,
    GenericClassMethod,
    frozen
)
from cocktail import schema
from cocktail.persistence import datastore, transaction
from woost.models import (
    Item,
    File,
    User,
    Slot,
    changeset_context
)


@enum.unique
class ExportMode(enum.Enum):
    ignore = 1
    export = 2
    expand = 3


@enum.unique
class UnknownMemberPolicy(enum.Enum):
    ignore = 1
    warn = 2
    report = 3
    fail = 4


@enum.unique
class MissingObjectPolicy(enum.Enum):
    ignore = 1
    warn = 2
    report = 3
    fail = 4
    create = 5


@GenericMethod
def get_object_ref(obj):

    ref = {
        "@class": get_full_name(obj.__class__)
    }

    for member in obj.__class__.iter_members():
        if (
            member is not Item.id
            and member.unique
            and not member.translated
            and isinstance(member, (schema.String, schema.Integer))
        ):
            value = obj.get(member)
            if value is not None and value != "":
                ref[member.name] = value

    return ref

@GenericClassMethod
def resolve_object_ref(cls, ref):

    for key, value in ref.iteritems():
        if value is not None and value != "" and not key.startswith("@"):
            member = cls.get_member(key)
            if member.unique and not member.translated:
                obj = cls.get_instance(**{key: value})
                if obj is not None:
                    return obj

    return None

@GenericClassMethod
def create_object_from_ref(cls, ref):
    obj = cls()

    for key, value in ref.iteritems():
        if not key.startswith("@"):
            member = obj.__class__.get_member(key)
            if member.unique and not member.translated:
                obj.set(key, value)

    obj.insert()
    return obj


class ObjectExporter(object):

    verbose = False
    language_subset = None
    date_format = "%Y-%m-%d"
    time_format = "%H:%M:%S"
    datetime_format = date_format + " " + time_format
    json_encoder_defaults = {"check_circular": False}

    def __init__(self):
        self.__member_export_modes = {
            Item.id: ExportMode.ignore,
            Item.translations: ExportMode.ignore,
            Item.last_update_time: ExportMode.ignore,
            Item.last_translation_update_time: ExportMode.ignore,
            Item.changes: ExportMode.ignore
        }
        self.__model_export_modes = TypeMapping()
        self.__exported_data = {}
        self.json_encoder_defaults = self.json_encoder_defaults.copy()

    def dump(self, dest, **kwargs):

        options = self.json_encoder_defaults.copy()
        options.update(kwargs)

        if isinstance(dest, basestring):
            with open(dest, "w") as file:
                json.dump(self.__exported_data.values(), file, **options)
        else:
            json.dump(self.__exported_data.values(), dest, **options)

    def dumps(self, **kwargs):
        options = self.json_encoder_defaults.copy()
        options.update(kwargs)
        return json.dumps(self.__exported_data.values(), **options)

    def add_all(self, objects):

        if self.verbose:
            if not isinstance(objects, Sequence):
                objects = list(objects)
            bar = ProgressBar(len(objects))
            bar.update()

        for obj in objects:
            self.add(obj)
            if self.verbose:
                bar.update(1)

        if self.verbose:
            bar.finish()

    def add(self, obj):

        if isinstance(obj, ExportNode):
            node = obj
            obj = node.value
        else:
            node = ExportNode(self, obj)
            node._export_mode = ExportMode.expand

        if obj in self.__exported_data:
            return False

        self.__exported_data[obj] = data = {
            "@class": get_full_name(node.value.__class__)
        }

        self._export_object_data(node, data)
        return True

    def _export_object_data(self, node, data):

        obj = node.value

        for member in obj.__class__.iter_members():
            if member.translated:
                translated_values = {}
                for lang in obj.translations:
                    member_node = ExportNode(
                        self,
                        obj.get(member, lang),
                        parent = node,
                        member = member,
                        language = lang
                    )
                    if member_node.export_mode != ExportMode.ignore:
                        translated_values[lang] = self._export_value(
                            member_node,
                            expand_object = (member_node.export_mode == ExportMode.expand)
                        )
                if translated_values:
                    data[member.name] = translated_values
            else:
                member_node = ExportNode(
                    self,
                    obj.get(member),
                    parent = node,
                    member = member
                )
                if member_node.export_mode != ExportMode.ignore:
                    data[member.name] = self._export_value(
                        member_node,
                        expand_object =
                            member_node.export_mode == ExportMode.expand
                    )

        # Encode file contents
        if isinstance(obj, File):
            buffer = StringIO()
            with open(obj.file_path, "rb") as file:
                base64.encode(file, buffer)
            buffer.seek(0)
            data["@file_data"] = buffer.getvalue()

    def _get_object_ref(self, node):
        return get_object_ref(node.value)

    def _export_value(self, node, expand_object = False):

        value = node.value

        if value is not None:
            if isinstance(value, type):
                value = get_full_name(value)
            elif isinstance(value, schema.SchemaObject):
                value = self._get_object_ref(node)
                if expand_object:
                    self.add(node)
            elif not isinstance(value, basestring) and isinstance(
                value,
                (
                    Sequence,
                    Set,
                    ListWrapper,
                    SetWrapper
                )
            ):
                items = []
                for index, item in enumerate(value):
                    item_node = ExportNode(
                        self,
                        item,
                        parent = node,
                        member = node.member.items,
                        language = node.language,
                        index = index
                    )
                    if item_node.export_mode != ExportMode.ignore:
                        item_copy = self._export_value(
                            item_node,
                            expand_object =
                                (item_node.export_mode == ExportMode.expand)
                        )
                        items.append(item_copy)
                value = items
            elif isinstance(value, datetime):
                return value.strftime(self.datetime_format)
            elif isinstance(value, date):
                return value.strftime(self.date_format)
            elif isinstance(value, time):
                return value.strftime(self.time_format)

        return value

    def get_node_export_mode(self, node):

        # Limit exported languages
        if (
            node.language
            and self.language_subset is not None
            and node.language not in self.language_subset
        ):
            return ExportMode.ignore

        if node.member:

            # Custom per member rules
            mode = self.get_member_export_mode(node.member)
            if mode:
                return mode

            # Exclude calculated fields
            if node.member.expression is not None:
                return ExportMode.ignore

            # Ignore anonymous relations
            if (
                isinstance(node.member, schema.RelationMember)
                and node.member.anonymous
            ):
                return ExportMode.ignore

        # Custom per model rules
        if isinstance(node.value, schema.SchemaObject):
            mode = self.get_model_export_mode(node.value.__class__)
            if mode:
                return mode

        # Standard member rules
        if node.member:

            # Always expand slots
            if isinstance(node.member, Slot):
                return ExportMode.expand

            # Always expand integral relations
            elif (
                isinstance(node.member, schema.RelationMember)
                and node.member.integral
            ):
                return ExportMode.expand

            # If the member is the item descriptor for a collection, apply the
            # behavior set by the collection
            if (
                node.parent
                and node.parent.member
                and isinstance(node.parent.member, schema.Collection)
                and node.parent.export_mode == ExportMode.expand
            ):
                return ExportMode.expand

        # If no other rule applies, assume a shallow copy
        return ExportMode.export

    def set_member_export_mode(self, member, mode):
        self.__member_export_modes[member] = mode

    def get_member_export_mode(self, member):
        return self.__member_export_modes.get(member)

    def set_model_export_mode(self, model, mode):
        self.__model_export_modes[model] = mode

    def get_model_export_mode(self, model):
        return self.__model_export_modes.get(model)

    def run_cli(self, func):
        parser = self._cli_create_parser()
        args = parser.parse_args()
        self._cli_process_args(args)
        func()
        self.dump(args.file)

    def _cli_create_parser(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "file",
            nargs = "?",
            type = argparse.FileType("w"),
            default = sys.stdout
        )
        parser.add_argument(
            "--verbose", "-v",
            action = "store_true",
            default = self.verbose
        )
        parser.add_argument(
            "--pretty-print", "-p",
            action = "store_true",
            default = False
        )
        return parser

    def _cli_process_args(self, args):

        self.verbose = args.verbose

        if args.pretty_print:
            self.json_encoder_defaults["indent"] = 2
            self.json_encoder_defaults["sort_keys"] = True


class ExportNode(object):

    _export_mode = None

    def __init__(self,
        exporter,
        value,
        parent = None,
        member = None,
        language = None,
        index = None
    ):
        self._exporter = exporter
        self._value = value
        self._parent = parent
        self._member = member
        self._language = language
        self._index = index

    def __repr__(self):

        props = []

        desc = "%s(%r" % (self.__class__.__name__, self._value)

        if self._parent:
            desc += ", parent=%r" % self._parent

        if self._member:
            desc += ", member=%r" % self._member

        if self._language:
            desc += ", language=%r" % self._language

        if self._index:
            desc += ", index=%r" % self._index

        return desc + ")"

    @property
    def exporter(self):
        return self._exporter

    @property
    def value(self):
        return self._value

    @property
    def schema_object(self):

        node = self

        while node is not None:
            if isinstance(node._value, schema.SchemaObject):
                return node._value
            node = node._parent

        return None

    @property
    def parent(self):
        return self._parent

    @property
    def member(self):
        return self._member

    @property
    def language(self):
        return self._language

    @property
    def index(self):
        return self._index

    @property
    def export_mode(self):
        if self._export_mode is None:
            self._export_mode = self._exporter.get_node_export_mode(self)
        return self._export_mode


class ObjectImporter(object):

    verbose = False
    unknown_member_policy = UnknownMemberPolicy.fail
    missing_object_policy = MissingObjectPolicy.fail
    language_subset = None
    date_format = ObjectExporter.date_format
    time_format = ObjectExporter.time_format
    datetime_format = ObjectExporter.datetime_format

    def __init__(self):
        self.__qname_mapping = {}
        self.missing_references = Counter()
        self.missing_members = defaultdict(Counter)

    def load(self, file, encoding = "utf-8"):

        if isinstance(file, basestring):
            with open(file, "r") as file:
                string = file.read()
        else:
            string = file.read()

        return self.loads(string, encoding)

    def loads(self, string, encoding = "utf-8"):
        if isinstance(string, str):
            data = string.decode(encoding)
        data = json.loads(string)
        return self.load_objects(data)

    def load_objects(self, object_list):

        # Make a first pass to pregenerate all new objects
        if self.verbose:
            print "Resolving or creating %d objects" % len(object_list)
            bar = ProgressBar(len(object_list))
            bar.update()

        objects = []

        for object_data in object_list:
            objects.append((
                self.resolve_object_ref(
                    object_data,
                    missing_object_policy = MissingObjectPolicy.create
                ),
                object_data
            ))
            if self.verbose:
                bar.update(1)

        if self.verbose:
            bar.finish()

        # And a second pass to load object state
        if self.verbose:
            print "Importing object state"
            bar = ProgressBar(len(objects))
            bar.update()

        for obj, object_data in objects:
            self.import_object_data(obj, object_data)
            if self.verbose:
                bar.update(1)

        if self.verbose:
            bar.finish()

        self.print_missing_references_report()
        self.print_missing_members_report()

    def print_missing_references_report(self):
        if self.missing_references:
            print "\nThe following references couldn't be resolved:\n"
            for ref, count in self.missing_references.most_common():
                print "  %s (%d)" % (ref, count)

    def print_missing_members_report(self):
        if self.missing_members:
            print "\nThe following members aren't defined:"
            for cls, keys in self.missing_members.iteritems():
                print "\n  %s:" % cls.__name__
                for key, count in keys.most_common():
                    print "  -%s (%d)" % (key, count)
            print

    def import_object_data(self, obj, data):

        for key, value in data.iteritems():

            if isinstance(key, unicode):
                key = str(key)

            if key.startswith("@"):

                # Store file contents once the current transaction is comitted
                # successfully
                if key == "@file_data":

                    datastore.unique_after_commit_hook(
                        "woost.models.objectio.store_file_contents",
                        _store_file_contents_after_commit
                    )

                    file_contents = datastore.get_transaction_value(
                        "woost.models.objectio.file_contents"
                    )

                    if file_contents is None:
                        file_contents = {}
                        datastore.set_transaction_value(
                            "woost.models.objectio.file_contents",
                            file_contents
                        )

                    bdata = base64.decodestring(value)
                    file_contents[obj.id] = bdata

                continue

            member = obj.__class__.get_member(key)
            if member is None:
                self._handle_unknown_member(obj, key)
                continue

            if member.translated:
                for lang, lang_value in value.iteritems():
                    if (
                        self.language_subset is None
                        or lang in self.language_subset
                    ):
                        imported_value = self.import_object_value(
                            obj,
                            member,
                            lang_value,
                            lang
                        )
                        if imported_value is not ExportMode.ignore:
                            obj.set(key, imported_value, lang)
            else:
                imported_value = self.import_object_value(obj, member, value)
                if imported_value is not ExportMode.ignore:
                    obj.set(key, imported_value)

    def _handle_unknown_member(self, obj, key):
        if self.unknown_member_policy == UnknownMemberPolicy.fail:
            raise ValueError(
                "The schema for %r doesn't define member %s"
                % (obj, key)
            )
        elif self.unknown_member_policy == UnknownMemberPolicy.warn:
            warn("The schema for %s doesn't define member %s" % (obj, key))
        elif self.unknown_member_policy == UnknownMemberPolicy.report:
            self.missing_members[obj.__class__][key] += 1
        elif self.unknown_member_policy == UnknownMemberPolicy.ignore:
            pass
        else:
            raise ValueError(
                "The schema for %r doesn't define member %s, and the "
                "object importer has set an invalid policy for "
                "dealing with unknown members (%r); it should be one "
                "of the values provided by the UnknownMemberPolicy "
                "enumeration"
                % (obj, key, self.unknown_member_policy)
            )

    def import_object_value(self, obj, member, value, language = None):
        if value is None:
            return None
        elif (
            not isinstance(value, basestring)
            and isinstance(value, (
                Sequence,
                Set,
                ListWrapper,
                SetWrapper
            ))
        ):
            return member.default_type(
                imported_item
                for imported_item in (
                    self.import_object_value(
                        obj,
                        member.items,
                        item
                    )
                    for item in value
                )
                if imported_item is not ExportMode.ignore
            )
        elif isinstance(member, schema.Reference):
            if member.class_family:
                return resolve(value)
            else:
                return self.resolve_object_ref(value)
        elif isinstance(member, schema.DateTime):
            return datetime.strptime(value, self.datetime_format)
        elif isinstance(member, schema.Date):
            return datetime.strptime(value, self.date_format).date()
        elif isinstance(member, schema.Time):
            return datetime.strptime(value, self.time_format).time()

        return value

    def resolve_object_ref(self, ref, missing_object_policy = None):

        cls = resolve(ref["@class"])
        obj = resolve_object_ref(cls, ref)

        if obj is None:
            obj = self._handle_missing_object(
                cls,
                ref,
                missing_object_policy or self.missing_object_policy
            )

        return obj

    def _handle_missing_object(self, cls, ref, policy):

        if policy == MissingObjectPolicy.create:

            if cls is None:
                raise ValueError(
                    "Can't create missing object %r; no class specified"
                    % ref
                )

            return create_object_from_ref(cls, ref)

        elif policy == MissingObjectPolicy.warn:
            warn("Can't find an object matching the reference %r" % ref)
            return ExportMode.ignore

        elif policy == MissingObjectPolicy.report:
            self.missing_references[frozen(ref)] += 1
            return ExportMode.ignore

        elif policy == MissingObjectPolicy.fail:
            raise ValueError(
                "Can't find an object matching the reference %r" % ref
            )

        elif policy == MissingObjectPolicy.ignore:
            return ExportMode.ignore

        else:
            raise ValueError(
                "Can't find an object matching the reference %r, and the "
                "object importer has set an invalid policy for dealing "
                "with missing objects (%r); expected one of the values of "
                "the MissingObjectPolicy enumeration"
                % (ref, policy)
            )

    def map_qname(self, original, replacement):
        self.__qname_mapping[original] = replacement

    def run_cli(self):
        parser = self._cli_create_parser()
        args = parser.parse_args()
        self._cli_process_args(args)

        @transaction
        def load_files():

            if args.author:
                author = User.require_instance(email = args.author)
            else:
                author = None

            with changeset_context(author):
                for file in args.files:
                    self.load(file)

    def _cli_create_parser(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "files",
            nargs = "+",
            type = argparse.FileType("r"),
            default = sys.stdout
        )
        parser.add_argument(
            "--verbose", "-v",
            action = "store_true",
            default = self.verbose
        )
        parser.add_argument(
            "--author"
        )
        parser.add_argument(
            "--errors", "-e"
        )
        parser.add_argument(
            "--missing-objects",
            choices = [
                choice
                for choice in MissingObjectPolicy.__members__
                if choice != "create"
            ]
        )
        parser.add_argument(
            "--unknown-members",
            choices = UnknownMemberPolicy.__members__.keys()
        )
        return parser

    def _cli_process_args(self, args):

        self.verbose = args.verbose

        missing_objects_key = args.missing_objects or args.errors
        if missing_objects_key:
            self.missing_object_policy = getattr(
                MissingObjectPolicy,
                missing_objects_key
            )

        unknown_members_key = args.unknown_members or args.errors
        if unknown_members_key:
            self.unknown_member_policy = getattr(
                UnknownMemberPolicy,
                unknown_members_key
            )


def _store_file_contents_after_commit(successful):

    if not successful:
        return

    file_contents = datastore.get_transaction_value(
        "woost.models.objectio.file_contents"
    )

    if file_contents is not None:

        for id, data in file_contents.iteritems():
            file = File.get_instance(id)
            if file is None:
                continue

            with open(file.file_path, "wb") as f:
                f.write(data)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import mimetypes
from cStringIO import StringIO
import cherrypy
from cocktail.translations import translations
from cocktail.schema.io import export_file
from cocktail.controllers import request_property
from woost.controllers.backoffice.useractions import get_user_action
from woost.controllers.backoffice.editcontroller import EditController


class SurveyResultsController(EditController):

    view_class = "woost.extensions.surveys.SurveyResultsView"

    allowed_rendering_formats = (
        EditController.allowed_rendering_formats
        | frozenset(["msexcel"])
    )

    @cherrypy.expose
    def render_msexcel(self):

        survey = self.edit_node.item
        survey_schema = survey.create_survey_schema()

        mime_type = mimetypes.types_map[".xls"]
        cherrypy.response.headers['Content-Type'] = mime_type
        cherrypy.response.headers["Content-Disposition"] = \
            'attachment; filename="%s"' % (translations(survey) + ".xls")

        buffer = StringIO()
        export_file(
            survey.submissions,
            buffer,
            survey_schema,
            mime_type = mime_type
        )
        return buffer.getvalue()

    @request_property
    def output(self):
        output = EditController.output(self)
        output["selected_action"] = get_user_action("survey_results")
        return output




#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			October 2008
"""

from woost import app
app.package = "--SETUP-PACKAGE--"
app.installation_id = "--SETUP-INSTALLATION_ID--"

# Application server configuration
import cherrypy
cherrypy.config.update({
    "global": {
        "server.socket_host": "--SETUP-APP_SERVER_HOSTNAME--",
        "server.socket_port": --SETUP-PORT--,
        "tools.encode.on": True,
        "tools.encode.encoding": "utf-8",
        "tools.decode.on": True,
        "tools.decode.encoding": 'utf-8'
    }
})

# Object store provider
from cocktail.persistence import datastore
from ZEO.ClientStorage import ClientStorage
db_host = "127.0.0.1"
db_port = --SETUP-ZEO_PORT--
datastore.storage = lambda: ClientStorage((db_host, db_port))

# Use file based sessions
from cocktail.controllers import session
session.config["session.type"] = "file"




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Item


class Variable(Item):

    members_order = [
        "identifier",
        "text",
        "translated_text",
        "dynamic_value"
    ]

    identifier = schema.String(
        required = True,
        unique = True,
        indexed = True,
        normalized_index = False,
        descriptive = True
    )

    text = schema.String(
        edit_control = "cocktail.html.TextArea"
    )

    translated_text = schema.String(
        translated = True,
        edit_control = "cocktail.html.TextArea"
    )

    dynamic_value = schema.CodeBlock(
        language = "python"
    )

    def get_value(self):

        text = self.translated_text or self.text

        expr = self.dynamic_value
        if expr:
            code_object = compile(
                self.dynamic_value,
                "%r.dynamic_value" % self,
                "exec"
            )
            context = {"self": self, "text": text}
            exec code_object in context
            text = context["text"]

        return text




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.translations import translations
from cocktail.html import templates
from woost.models import Extension, extension_translations

translations.define("TranslationWorkflowExtension",
    ca = u"Circuït de traducció",
    es = u"Circuito de traducción",
    en = u"Translation workflow"
)

translations.define("TranslationWorkflowExtension-plural",
    ca = u"Circuït de traducció",
    es = u"Circuito de traducción",
    en = u"Translation workflow"
)


class TranslationWorkflowExtension(Extension):

    standard_states = (
        {"id": "pending", "color": "#c71e1e"},
        {"id": "ignored", "color": "#666666"},
        {"id": "silenced", "color": "#333333"},
        {"id": "selected", "color": "#e37e20"},
        {"id": "in_translation", "color": "#e6ce18"},
        {"id": "proposed", "color": "#189627"},
        {"id": "applied", "color": "#ffffff"}
    )

    # transition_id: (source_states, target_state)
    standard_transitions = (
        ("ignore", {
            "source_states": [],
            "target_state": "ignored",
            "icon": "ignore"
        }),
        ("silence", {
            "source_states": [],
            "target_state": "silenced",
            "icon": "silence"
        }),
        ("select", {
            "source_states": ["pending", "ignored", "silenced"],
            "target_state": "selected",
            "icon": "forward"
        }),
        ("reject_selection", {
            "source_states": ["selected"],
            "target_state": "pending",
            "icon": "backward"
        }),
        ("assign", {
            "source_states": ["selected", "applied"],
            "target_state": "in_translation",
            "icon": "forward",
            "setup": "woost.extensions.translationworkflow."
                     "assigntransitionsetup."
                     "TranslationWorkflowAssignTransitionSetup",
            "code":
                "from woost.extensions.translationworkflow.translationagency "
                "import TranslationAgency\n"
                "path = request.source_language, request.target_language\n"
                "translator = data['assignments'][path]\n"
                "if isinstance(translator, TranslationAgency):\n"
                "    request.assigned_agency = translator\n"
                "else:\n"
                "    request.assigned_translator = translator"
        }),
        ("propose", {
            "source_states": ["in_translation"],
            "target_state": "proposed",
            "icon": "forward",
            "requires_valid_text": True
        }),
        ("cancel_proposal", {
            "source_states": ["proposed"],
            "target_state": "in_translation",
            "icon": "backward"
        }),
        ("reject", {
            "source_states": ["proposed"],
            "target_state": "in_translation",
            "icon": "backward"
        }),
        ("accept", {
            "source_states": ["proposed"],
            "target_state": "applied",
            "icon": "forward",
            "code": "request.apply_translated_values()"
        })
    )

    def __init__(self, **values):
        Extension.__init__(self, **values)
        self.extension_author = u"Whads/Accent SL"
        self.set("description",
            u"""Gestió avançada de traduccions.""",
            "ca"
        )
        self.set("description",
            u"""Gestión avanzada de traducciones.""",
            "es"
        )
        self.set("description",
            u"""Advanced translation management.""",
            "en"
        )

    def _load(self):
        from woost.extensions.translationworkflow import (
            strings,
            typegroups,
            configuration,
            user,
            item,
            role,
            request,
            comment,
            state,
            transition,
            transitionpermission,
            transitionaction,
        )
        from woost.extensions.translationworkflow.translationagency \
            import TranslationAgency

        # Disable user actions for translation requests
        from woost.controllers.backoffice.useractions import get_user_action
        from .request import TranslationWorkflowRequest

        get_user_action("delete").excluded_content_types.add(
            TranslationWorkflowRequest
        )

        # Register and disable the user actions for each transition
        from cocktail.events import when
        from woost.controllers.backoffice.useractions import get_user_actions
        from woost.controllers.backoffice.basebackofficecontroller \
            import BaseBackOfficeController
        from .transition import TranslationWorkflowTransition
        from .transitionaction \
            import TranslationWorkflowTransitionAction as TWTransAction

        @when(BaseBackOfficeController.before_request)
        def register_transition_user_actions(e):
            for action in get_user_actions():
                if (
                    isinstance(action, TWTransAction)
                    and action.transition is None
                ):
                    action.enabled = False

            for transition in TranslationWorkflowTransition.select(
                order = "relative_order"
            ):
                TWTransAction.register_transition_action(transition)

        # Install the transition controller
        from woost.controllers.backoffice.backofficecontroller \
            import BackOfficeController
        from .transitioncontroller \
            import TranslationWorkflowTransitionController

        BackOfficeController.translation_workflow_transition = \
            TranslationWorkflowTransitionController

        # Install overlays
        from cocktail.html import templates
        for overlay_id in (
            "BackOfficeLayout",
            "BackOfficeFieldsView",
            "ContentView",
            "TranslationDisplay"
        ):
            templates.get_class(
                "woost.extensions.translationworkflow.%sOverlay"
                % overlay_id
            )

        # Optional automatic transition on edit controllers
        from woost.controllers.backoffice.editcontroller import EditController
        from woost.extensions.translationworkflow.defaulttransition \
            import apply_default_transition
        EditController.saving_item.append(apply_default_transition)

        # Count affected translation requests after editing an item
        from cocktail.events import when
        from woost.controllers.backoffice.editstack import EditNode
        from woost.extensions.translationworkflow.utils \
            import notify_translation_request_changes

        @when(EditNode.committed)
        def show_edit_effects_on_translation_workflow(e):
            if e.changeset is not None:
                notify_translation_request_changes(e.changeset)

        # When the manager of a translation agency creates a user, set it up as
        # a translator for the agency automatically
        from woost.models import User, Role
        from woost.controllers.backoffice.editstack import EditNode

        @when(EditNode.saving)
        def link_to_translation_agency(e):

            if e.is_new and isinstance(e.item, User):
                agency = e.user.managed_translation_agency
                if agency is not None:

                    # Register a user as a translator for the agency
                    if e.item not in agency.translators:
                        agency.translators.append(e.item)

                    # Grant the user translator rights
                    translator_role = Role.get_instance(
                        qname = "woost.extensions.translationworkflow."
                                "roles.translators"
                    )
                    if translator_role is not None and not e.item.roles:
                        e.item.roles.append(translator_role)

    def _install(self):
        self.create_standard_graph()
        self.create_standard_access_restrictions()

    def create_standard_graph(self):

        from pkg_resources import resource_filename
        from woost.models import File
        from woost.extensions.translationworkflow import installationstrings
        from woost.extensions.translationworkflow.state \
            import TranslationWorkflowState
        from woost.extensions.translationworkflow.transition \
            import TranslationWorkflowTransition

        # Create all the possible states
        states = {}

        for state_data in self.standard_states:
            state_id = state_data["id"]
            state = self._create_asset(
                TranslationWorkflowState,
                "states." + state_id,
                title = extension_translations,
                plural_title = extension_translations,
                color = state_data["color"]
            )
            states[state_id] = state

        # Setup the state change that should be applied when the source object
        # for a translation request is modified (by default, move the request
        # back to the 'pending' state)
        pending_state = states["pending"]
        silenced_state = states["silenced"]
        for state in states.itervalues():
            if state not in (pending_state, silenced_state):
                state.state_after_source_change = pending_state

        # Create the transitions between states
        rel_order = 0
        icons = {}

        for transition_id, transition_data in self.standard_transitions:

            icon_id = transition_data["icon"]
            icon = icons.get(icon_id)
            if icon is None:
                icons[icon_id] = icon = self._create_asset(
                    lambda:
                        File.from_path(
                            resource_filename(
                                "woost.extensions.translationworkflow",
                                "icons/%s.png" % icon_id
                            ),
                            languages = ()
                        ),
                    "icons." + icon_id,
                    title = extension_translations
                )

            rel_order += 10

            self._create_asset(
                TranslationWorkflowTransition,
                "transitions." + transition_id,
                title = extension_translations,
                source_states = map(
                    states.get,
                    transition_data["source_states"]
                ),
                target_state = states[transition_data["target_state"]],
                relative_order = rel_order,
                transition_setup_class = transition_data.get("setup"),
                transition_code = transition_data.get("code"),
                requires_valid_text = transition_data.get(
                    "requires_valid_text",
                    False
                ),
                icon = icon
            )

    def create_standard_access_restrictions(self):

        from woost.models import (
            Role,
            ReadPermission,
            CreatePermission,
            ModifyPermission,
            DeletePermission,
            ReadMemberPermission
        )
        from woost.extensions.translationworkflow import installationstrings
        from woost.extensions.translationworkflow.request \
            import TranslationWorkflowRequest
        from woost.extensions.translationworkflow.comment \
            import TranslationWorkflowComment
        from woost.extensions.translationworkflow.state \
            import TranslationWorkflowState
        from woost.extensions.translationworkflow.transition \
            import TranslationWorkflowTransition
        from woost.extensions.translationworkflow.transitionpermission \
            import TranslationWorkflowTransitionPermission

        prefix = "woost.extensions.translationworkflow."

        def get_items(cls, *args):
            return [cls.require_instance(qname = prefix + arg) for arg in args]

        # Administrators
        administrators = Role.require_instance(qname = "woost.administrators")
        administrators.permissions.append(
            self._create_asset(
                TranslationWorkflowTransitionPermission,
                "permissions.administrators.transition"
            )
        )

        # Editors
        editors = Role.require_instance(qname = "woost.editors")
        editors.translation_workflow_relevant_states = get_items(
            TranslationWorkflowState,
            "states.pending",
            "states.ignored",
            "states.silenced",
            "states.selected",
            "states.proposed",
            "states.applied"
        )
        editors.translation_workflow_default_state = (
            TranslationWorkflowState.require_instance(
                qname = prefix + "states.pending"
            )
        )
        editors.translation_workflow_default_transition = (
            TranslationWorkflowTransition.require_instance(
                qname = prefix + "transitions.select"
            )
        )
        editors.translation_workflow_default_transition_condition = (
            "applies = (request.state.qname == "
            "'woost.extensions.translationworkflow.states.pending')"
        )
        editors.permissions.append(
            # Allow editors to execute a subset of transitions
            self._create_asset(
                TranslationWorkflowTransitionPermission,
                "permissions.editors.transition",
                transitions = get_items(
                    TranslationWorkflowTransition,
                    "transitions.ignore",
                    "transitions.silence",
                    "transitions.select",
                    "transitions.reject",
                    "transitions.accept"
                )
            )
        )

        # Coordinators
        coordinators = self._create_asset(
            Role,
            "roles.coordinators",
            title = extension_translations,
            default_content_type = TranslationWorkflowRequest,
            translation_workflow_relevant_states = get_items(
                TranslationWorkflowState,
                "states.selected",
                "states.in_translation",
                "states.proposed",
                "states.applied"
            ),
            translation_workflow_default_state = (
                TranslationWorkflowState.require_instance(
                    qname = prefix + "states.selected"
                )
            ),
            permissions = [
                # Allow coordinators to execute a subset of transitions
                self._create_asset(
                    TranslationWorkflowTransitionPermission,
                    "permissions.coordinators.transition",
                    transitions = get_items(
                        TranslationWorkflowTransition,
                        "transitions.reject_selection",
                        "transitions.assign"
                    )
                ),
                # See all requests
                self._create_asset(
                    ReadPermission,
                    "permissions.coordinators.read_requests",
                    content_type = TranslationWorkflowRequest
                ),
                # Read comments
                self._create_asset(
                    ReadPermission,
                    "permissions.coordinators.read_comments",
                    content_type = TranslationWorkflowComment
                ),
                # Make comments
                self._create_asset(
                    CreatePermission,
                    "permissions.coordinators.write_comments",
                    content_type = TranslationWorkflowComment
                ),
                # Limit the members they can see
                self._create_asset(
                    ReadMemberPermission,
                    "permissions.coordinators.read_members",
                    matching_members =
                        [
                            "woost.models.item.Item." + k
                            for k in [
                                "id",
                                "creation_time",
                                "last_update_time",
                                "author"
                            ]
                        ]
                        + [
                            prefix + "request.TranslationWorkflowRequest." + k
                            for k in [
                                "translated_item",
                                "source_language",
                                "target_language",
                                "state",
                                "assigned_translator",
                                "comments",
                                "translated_values"
                            ]
                        ]
                ),
                # Deny everything else
                self._create_asset(
                    CreatePermission,
                    "permissions.coordinators.deny_create",
                    authorized = False
                ),
                self._create_asset(
                    ModifyPermission,
                    "permissions.coordinators.deny_modify",
                    authorized = False
                ),
                self._create_asset(
                    DeletePermission,
                    "permissions.coordinators.deny_delete",
                    authorized = False
                ),
                self._create_asset(
                    ReadMemberPermission,
                    "permissions.coordinators.deny_read_members",
                    authorized = False
                )
            ]
        )

        # Translators
        translators = self._create_asset(
            Role,
            "roles.translators",
            title = extension_translations,
            default_content_type = TranslationWorkflowRequest,
            translation_workflow_relevant_states = get_items(
                TranslationWorkflowState,
                "states.in_translation",
                "states.proposed",
                "states.applied"
            ),
            translation_workflow_default_state = (
                TranslationWorkflowState.require_instance(
                    qname = prefix + "states.in_translation"
                )
            ),
            permissions = [
                # Allow translators to execute a subset of transitions
                self._create_asset(
                    TranslationWorkflowTransitionPermission,
                    "permissions.translators.transition",
                    transitions = get_items(
                        TranslationWorkflowTransition,
                        "transitions.propose",
                        "transitions.cancel_proposal"
                    )
                ),
                # Allow translators to see their assigned requests
                self._create_asset(
                    ReadPermission,
                    "permissions.translators.read_requests",
                    subject_description = extension_translations,
                    content_type = TranslationWorkflowRequest,
                    content_expression =
                        "items.add_filter(cls.assigned_translator.equal(user))"
                ),
                # Allow translators to modify their assigned translation
                # requests
                self._create_asset(
                    ModifyPermission,
                    "permissions.translators.modify_assigned_requests",
                    subject_description = extension_translations,
                    content_type = TranslationWorkflowRequest,
                    content_expression =
                        "from woost.extensions.translationworkflow.state \\\n"
                        "    import TranslationWorkflowState\n"
                        "in_translation = TranslationWorkflowState.require_instance(\n"
                        "    qname = 'woost.extensions.translationworkflow.states.in_translation'\n"
                        ")\n"
                        "items.add_filter(cls.state.equal(in_translation))\n"
                        "items.add_filter(cls.assigned_translator.equal(user))"
                ),
                # Allow translators to see comments for their assigned requests
                self._create_asset(
                    ReadPermission,
                    "permissions.translators.read_comments",
                    subject_description = extension_translations,
                    content_type = TranslationWorkflowComment,
                    content_expression =
                        "from woost.extensions.translationworkflow.request \\\n"
                        "    import TranslationWorkflowRequest as TWRequest\n"
                        "assigned_requests = TWRequest.select(\n"
                        "    TWRequest.assigned_translator.equal(user)\n"
                        ")\n"
                        "items.add_filter(cls.request.one_of(assigned_requests))"
                ),
                # Comment requests assigned to them
                self._create_asset(
                    CreatePermission,
                    "permissions.translators.write_comments",
                    subject_description = extension_translations,
                    content_type = TranslationWorkflowComment,
                    content_expression =
                        "from woost.extensions.translationworkflow.request \\\n"
                        "    import TranslationWorkflowRequest as TWRequest\n"
                        "assigned_requests = TWRequest.select(\n"
                        "    TWRequest.assigned_translator.equal(user)\n"
                        ")\n"
                        "items.add_filter(cls.request.one_of(assigned_requests))"
                ),
                # Limit the members they can see
                self._create_asset(
                    ReadMemberPermission,
                    "permissions.translators.read_members",
                    matching_members =
                        [
                            "woost.models.item.Item." + k
                            for k in [
                                "id",
                                "creation_time",
                                "last_update_time",
                                "author"
                            ]
                        ]
                        + [
                            prefix + "request.TranslationWorkflowRequest." + k
                            for k in [
                                "translated_item",
                                "source_language",
                                "target_language",
                                "state",
                                "comments",
                                "translated_values"
                            ]
                        ]
                ),
                # Deny everything else
                self._create_asset(
                    CreatePermission,
                    "permissions.translators.deny_create",
                    authorized = False
                ),
                self._create_asset(
                    ModifyPermission,
                    "permissions.translators.deny_modify",
                    authorized = False
                ),
                self._create_asset(
                    DeletePermission,
                    "permissions.translators.deny_delete",
                    authorized = False
                ),
                self._create_asset(
                    ReadMemberPermission,
                    "permissions.translators.deny_read_members",
                    authorized = False
                )
            ]
        )




#-*- coding: utf-8 -*-
u"""Defines migrations to the database schema for woost.

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.events import when
from cocktail.persistence import MigrationStep
from cocktail.persistence.migration import migration_steps
from cocktail.persistence.utils import remove_broken_type
from warnings import warn

def admin_members_restriction(members):

    def add_permission(e):

        from woost.models import Role, ModifyMemberPermission

        everybody_role = Role.require_instance(qname = "woost.everybody")
        permission = ModifyMemberPermission(
            matching_members = list(members),
            authorized = False
        )
        permission.insert()

        for i, p in enumerate(everybody_role.permissions):
            if isinstance(p, ModifyMemberPermission) and p.authorized:
                everybody_role.permissions.insert(i, permission)
                break
        else:
            everybody_role.permissions.append(permission)

    return add_permission

#------------------------------------------------------------------------------

step = MigrationStep("added woost.models.Document.robots_*")

step.executing.append(
    admin_members_restriction([
        "woost.models.document.Document.robots_should_index",
        "woost.models.document.Document.robots_should_follow"
    ])
)

@step.processor("woost.models.document.Document")
def set_defaults(document):
    if not hasattr(document, "_robots_should_index"):
        document.robots_should_index = True
        document.robots_should_follow = True

#------------------------------------------------------------------------------

step = MigrationStep("added woost.models.Publishable.requires_https")

step.executing.append(
    admin_members_restriction([
        "woost.models.publishable.Publishable.requires_https"
    ])
)

@step.processor("woost.models.publishable.Publishable")
def set_defaults(publishable):
    if not hasattr(publishable, "_requires_https"):
        publishable.requires_https = False


#------------------------------------------------------------------------------

step = MigrationStep("make Product extend Publishable")

@when(step.executing)
def update_keys(e):
    from woost.extensions.shop import ShopExtension

    if ShopExtension.enabled:
        from cocktail.translations import translations
        from woost.models import Publishable
        from woost.extensions.shop import create_product_controller
        from woost.extensions.shop.product import Product

        # Update the publishable keys
        Publishable.keys.update([product.id for product in Product.select()])

        # Create the product controller
        create_product_controller()

#------------------------------------------------------------------------------

step = MigrationStep("use TranslationMapping to translations")

@when(step.executing)
def update_translations(e):
    from cocktail.schema import TranslationMapping
    from cocktail.persistence import PersistentObject

    def translated_items(schema):
        if schema.translated and schema.indexed:
            for item in schema.select():
                yield item
        else:
            for derived_schema in schema.derived_schemas(False):
                for item in translated_items(derived_schema):
                    yield item

    for item in translated_items(PersistentObject):
        translations = TranslationMapping(
            owner = item,
            items = item.translations._items
        )
        item.translations._items = translations
        item._p_changed = True

    PersistentObject.rebuild_indexes(True)

#------------------------------------------------------------------------------

step = MigrationStep(
    "rename EmailTemplate.embeded_images to EmailTemplate.attachments"
)

step.rename_member(
    "woost.models.EmailTemplate",
    "embeded_images",
    "attachments"
)

#------------------------------------------------------------------------------

step = MigrationStep(
    "Apply full text indexing to elements with no translations"
)

@when(step.executing)
def rebuild_full_text_index(e):
    from woost.models import Item
    Item.rebuild_full_text_indexes(True)

#------------------------------------------------------------------------------

step = MigrationStep(
    "Replace EmailTemplate.attachments with EmailTemplate.initialization_code"
)

@when(step.executing)
def relocate_attachments_code(e):
    from woost.models import EmailTemplate

    for email_template in EmailTemplate.select():
        code = getattr(email_template, "_attachments", None)
        if code:
            del email_template._attachments
            if email_template.initialization_code:
                code = email_template.initialization_code + "\n\n" + code
            email_template.initialization_code = code

#------------------------------------------------------------------------------

step = MigrationStep(
    "Added the Role.implicit member"
)

@when(step.executing)
def flag_implicit_roles(e):
    from woost.models import Role

    implicit_roles_qnames = set((
        "woost.anonymous",
        "woost.everybody",
        "woost.authenticated"
    ))

    for role in Role.select():
        role.implicit = (role.qname in implicit_roles_qnames)

#------------------------------------------------------------------------------

step = MigrationStep("Removed the File.local_path member")

@step.processor("woost.models.file.File")
def remove_file_local_path_values(file):
    try:
        del file._local_path
    except:
        pass

@step.processor("woost.models.permission.MemberPermission")
def remove_permissions_for_file_local_path(permission):
    try:
        permission.matching_members.remove("woost.models.file.File.local_path")
    except:
        pass

#------------------------------------------------------------------------------

step = MigrationStep("Added the Style.applicable_to_text member")

@when(step.executing)
def index_style_applicable_to_text_member(e):
    from woost.models import Style
    Style.applicable_to_text.rebuild_index()

#------------------------------------------------------------------------------

step = MigrationStep(
    "Replaced CachingPolicy.cache_expiration with "
    "CachingPolicy.expiration_expression"
)

@step.processor("woost.models.caching.CachingPolicy")
def replace_cache_expiration_with_expiration_expression(policy):
    expiration = getattr(policy, "_cache_expiration", None)
    if expiration is not None:
        policy.expiration_expression = "expiration = %s" % expiration
        del policy._cache_expiration

#------------------------------------------------------------------------------

step = MigrationStep("Add multisite support")

@when(step.executing)
def add_multisite_support(e):
    from cocktail.persistence import datastore
    from woost.models import Configuration, Website, Item
    root = datastore.root

    # Remove all back-references from the Site and Language models
    for item in Item.select():
        for key in dir(item):
            if (
                key == "_site"
                or key.startswith("_Site_")
                or key.startswith("_Language_")
            ):
                delattr(item, key)

    # Remove the instance of Site from the database
    site_id = list(Item.qname.index.values(key = "woost.main_site"))[0]
    site = Item.index[site_id]
    site_state = site.__Broken_state__.copy()
    site_state["translations"] = dict(
        (lang, translation.__Broken_state__.copy())
        for lang, translation in site_state.pop("_translations").iteritems()
    )
    Item.index.remove(site_id)
    Item.keys.remove(site_id)

    # Create the configuration object
    config = Configuration()
    config.qname = "woost.configuration"
    config.insert()

    # Create a website
    website = Website()
    website.insert()
    website.hosts = ["localhost"]
    config.websites.append(website)

    # Languages
    published_languages = []

    for lang_id in root["woost.models.language.Language-keys"]:
        language = Item.index[lang_id]
        Item.index.remove(lang_id)
        Item.keys.remove(lang_id)
        language_state = language.__Broken_state__
        config.languages.append(language_state["_iso_code"])
        if language_state["_enabled"]:
            published_languages.append(language_state["_iso_code"])

    if list(config.languages) != published_languages:
        config.published_languages = published_languages

    # Delete old indexes from the database
    for key in list(root):
        if (
            key.startswith("woost.models.site.Site")
            or key.startswith("woost.models.language.Language")
        ):
            del root[key]

    # Settings that now belong in Configuration, as attributes
    config.secret_key = site_state.pop("secret_key")

    # Settings that now belong in Configuration, as regular fields
    for key in (
        "login_page",
        "generic_error_page",
        "not_found_error_page",
        "forbidden_error_page",
        "default_language",
        "backoffice_language",
        "heed_client_language",
        "timezone",
        "smtp_host",
        "smtp_user",
        "smtp_password"
    ):
        config.set(key, site_state.pop("_" + key))

    # Settings that now belong in Configuration, as collections
    for key in (
        "publication_schemes",
        "caching_policies",
        "renderers",
        "image_factories",
        "triggers"
    ):
        config.set(key, list(site_state.pop("_" + key)))

    # Settings that now belong in Website, becoming translated fields
    for key in (
        "town",
        "region",
        "country"
    ):
        value = site_state.pop("_" + key)
        for lang in config.languages:
            website.set(key, value, lang)

    # Settings that now belong in website, as translated fields
    for key in (
        "site_name",
        "organization_name",
        "keywords",
        "description"
    ):
        for lang, translation_state in site_state["translations"].iteritems():
            value = translation_state.pop("_" + key)
            website.set(key, value, lang)

    # Settings that now belong in website, as regular fields
    for key in (
        "logo",
        "icon",
        "home",
        "organization_url",
        "address",
        "postal_code",
        "phone_number",
        "fax_number",
        "email",
        "https_policy",
        "https_persistence"
    ):
        website.set(key, site_state.pop("_" + key))

    # Extension specific changes
    from woost.extensions.blocks import BlocksExtension
    if BlocksExtension.instance.enabled:
        config.common_blocks = list(site_state.pop("_common_blocks"))

    from woost.extensions.audio import AudioExtension
    if AudioExtension.instance.enabled:
        config.audio_encoders = list(site_state.pop("_audio_encoders"))
        config.audio_decoders = list(site_state.pop("_audio_decoders"))

    from woost.extensions.mailer import MailerExtension
    if MailerExtension.instance.enabled:
        from woost.extensions.mailer.mailing import Mailing
        for mailing in Mailing.select():
            language = mailing._language
            if language:
                mailing._language = language.__Broken_state__["_iso_code"]

    from woost.extensions.googleanalytics import GoogleAnalyticsExtension
    if GoogleAnalyticsExtension.instance.enabled:
        account = GoogleAnalyticsExtension.instance._account
        del GoogleAnalyticsExtension.instance._account
        config.google_analytics_account = account

    # Rebuild all indexes
    Item.rebuild_indexes()

    # Preserve the remaining state
    datastore.root["woost.models.migration.multisite_leftovers"] = site_state

#------------------------------------------------------------------------------
step = MigrationStep("Store hashes using hexadecimal characters")

@when(step.executing)
def transform_hashes(e):
    from woost.models import File, User
    to_hex_string = lambda s: "".join(("%x" % ord(c)).zfill(2) for c in s)

    for file in File.select():
        file._file_hash = to_hex_string(file.file_hash)

    for user in User.select():
        if user.password:
            user._password = to_hex_string(user.password)

#------------------------------------------------------------------------------
step = MigrationStep("Assign global object identifiers")

@when(step.executing)
def assign_global_identifiers(e):
    from woost import app
    from woost.models import Item
    from woost.models.synchronization import rebuild_manifest

    for item in Item.select():
        item._global_id = app.installation_id + "-" + str(item.id)

    Item.global_id.rebuild_index()
    Item.synchronizable.rebuild_index()
    rebuild_manifest(True)

#------------------------------------------------------------------------------
step = MigrationStep("Expose models that where hidden in the Configuration model")

@when(step.executing)
def expose_hidden_configuration(e):
    from woost.models import Configuration

    config = Configuration.instance

    # restrictedaccess extension
    access_restrictions = getattr(config, "_access_restrictions", None)
    if access_restrictions:
        for restriction in access_restriction:
            try:
                del restriction._Configuration_access_restrictions
            except AttributeError:
                pass
        try:
            del config._access_restrictions
        except:
            pass

#------------------------------------------------------------------------------

step = MigrationStep("Remove the Action model")

@when(step.executing)
def remove_action_model(e):

    from cocktail.persistence import datastore
    from woost.models import Change

    root = datastore.root
    root.pop("woost.models.action.Action-keys", None)
    root.pop("woost.models.action.Action.id", None)
    root.pop("woost.models.action.Action.identifier", None)
    root.pop("woost.models.action.Action.title", None)

    for change in Change.select():
        change.action = change.action.__Broken_state__["_identifier"]

#------------------------------------------------------------------------------

step = MigrationStep("Remove the workflow extension")

@when(step.executing)
def remove_workflow_extension(e):

    from cocktail.persistence import datastore
    from woost.models import Item, Extension

    for extension in Extension.select():
        bp = getattr(extension, "__Broken_Persistent__", None)

        if bp is not None and bp.__name__ == "WorkflowExtension":
            id = extension.__Broken_state__["_id"]

            try:
                Item.index.remove(id, None)
            except KeyError:
                pass

            try:
                Item.keys.remove(id)
            except KeyError:
                pass

            try:
                Extension.keys.remove(id)
            except KeyError:
                pass

            break

            for member in Extension.members().itervalues():
                if member.indexed:
                    member.rebuild_index()

    for key in list(datastore.root):
        if key.startswith("woost.extensions.workflow"):
            datastore.root.pop(key)

#------------------------------------------------------------------------------

step = MigrationStep("Removed drafts")

@when(step.executing)
def remove_drafts(e):

    from cocktail.persistence import datastore
    from woost.models import (
        Item,
        Permission,
        ContentPermission,
        MemberPermission,
        Trigger,
        ContentTrigger
    )
    from woost.models.utils import remove_broken_type

    remove_broken_type(
        "woost.models.permission.ConfirmDraftPermission",
        existing_bases = (
            Item,
            Permission,
            ContentPermission
        )
    )

    remove_broken_type(
        "woost.models.permission.ConfirmDraftPermission",
        existing_bases = (
            Item,
            Trigger,
            ContentTrigger
        )
    )

    for item in Item.select():
        for key in "_is_draft", "_draft_source", "_drafts":
            try:
                delattr(item, key)
            except AttributeError:
                pass

    datastore.root.pop("woost.models.item.Item.is_draft", None)
    datastore.root.pop("woost.models.item.Item.draft_source", None)

    for permission in MemberPermission.select():
        for key in ("drafts", "draft_source", "is_draft"):
            key = "woost.models.item.Item." + key
            if key in permission.matching_members:
                permission.matching_members.remove(key)

#------------------------------------------------------------------------------

step = MigrationStep(
    "Change the qualified name of the password reset email template"
)

@when(step.executing)
def change_qname_of_password_reset_email_template(e):

    from woost.models import EmailTemplate

    template = EmailTemplate.get_instance(
        qname = "woost.views.password_change_confirmation_email_template"
    )

    if template:
        template.qname = "woost.password_change_email_template"

#------------------------------------------------------------------------------

step = MigrationStep("Move the blocks extension into the woost core")

@when(step.executing)
def move_blocks_to_core(e):

    from cocktail.persistence import datastore
    from woost.models import Item, Page, Template, Block, TextBlock
    from woost.models.rendering import Renderer

    # Rename keys
    for key in list(datastore.root):
        if key.startswith("woost.extensions.blocks."):
            new_key = key.replace(".extensions.blocks.", ".models.")
            datastore.root[new_key] = datastore.root.pop(key)

    # Consolidate StandardPage and BlocksPage into the new Page model
    for block in Block.select():
        try:
            pages = block._BlocksPage_blocks
        except AttributeError:
            pass
        else:
            block._Page_blocks = pages
            del block._BlocksPage_blocks

    page_ids = datastore.root.pop("woost.models.standardpage.StandardPage-keys")

    if page_ids:
        for id in page_ids:
            Page.keys.add(id)
            page = Page.get_instance(id)
            page._p_changed = True

            block = TextBlock()

            for lang, trans in page.translations.iteritems():
                if hasattr(trans, "_body"):
                    block.set("text", trans._body, lang)
                    del trans._body

            page.blocks.append(block)
            block.insert()

    page_ids = datastore.root.pop("woost.models.blockspage.BlocksPage-keys")

    if page_ids:
        for id in page_ids:
            Page.keys.add(id)

    # Update references to the old template for block pages
    for template in Template.select():
        if template.identifier == "woost.extensions.blocks.BlocksPageView":
            template.identifier = "woost.views.StandardView"

    # Fix the name in the related end for Page.blocks
    for block in Block.select():
        block.Page_blocks._PersistentRelationCollection__member_name = \
            "Page_blocks"

#------------------------------------------------------------------------------

step = MigrationStep("Remove the Item.owner field")

@when(step.executing)
def remove_owner_field(e):

    from cocktail.persistence import datastore
    from woost.models import Item, ContentPermission, MemberPermission

    # Remove the owner value of every item
    for item in Item.select():
        try:
            del item._owner
        except AttributeError:
            pass

    # Drop the index for the member
    full_member_name = "woost.models.item.Item.owner"
    datastore.root.pop(full_member_name, None)

    # Purge all references to the owner member from member permissions
    for permission in MemberPermission.select():
        if permission.matching_members:
            member_count = len(permission.matching_members)
            try:
                permission.matching_members.remove(full_member_name)
            except (KeyError, ValueError):
                pass
            else:
                if member_count == 1:
                    permission.delete()

    # Purge the 'owned-items' expression from all permissions
    for permission in ContentPermission.select():
        matching_items = permission.matching_items
        if matching_items:
            try:
                filter = permission.matching_items["filter"]
            except KeyError:
                pass
            else:
                if filter == "owned-items" or (
                    hasattr(filter, "__contains__")
                    and "owned-items" in filter
                ):
                    permission.delete()

#------------------------------------------------------------------------------

step = MigrationStep("Remove the Document.attachments member")

@when(step.executing)
def remove_attachments_member(e):

    from woost.models import Document, File, MemberPermission

    for document in Document.select():
        try:
            del document._attachments
        except AttributeError:
            pass

    for file in File.select():
        try:
            del file._Document_attachments
        except AttributeError:
            pass

    full_member_name = "woost.models.document.Document.attachment"

    for permission in MemberPermission.select():
        if permission.matching_members:
            member_count = len(permission.matching_members)
            try:
                permission.matching_members.remove(full_member_name)
            except (KeyError, ValueError):
                pass
            else:
                if member_count == 1:
                    permission.delete()

#------------------------------------------------------------------------------

step = MigrationStep("Remove per document resources")

@when(step.executing)
def remove_document_resources(e):

    from woost.models import Publishable, Document, MemberPermission

    for document in Document.select():

        try:
            del document._branch_resources
        except AttributeError:
            pass

        try:
            del document._page_resources
        except AttributeError:
            pass

    for publishable in Publishable.select():

        try:
            del publishable._Document_branch_resources
        except AttributeError:
            pass

        try:
            del publishable._Document_page_resources
        except AttributeError:
            pass

        try:
            del publishable._inherit_resources
        except AttributeError:
            pass

    members = (
        "woost.models.document.Document.branch_resources",
        "woost.models.document.Document.page_resources",
        "woost.models.publishable.Publishable.inherit_resources"
    )

    for permission in MemberPermission.select():
        if permission.matching_members:
            removed = False

            for full_member_name in members:
                try:
                    permission.matching_members.remove(full_member_name)
                except (KeyError, ValueError):
                    pass
                else:
                    removed = True

            if removed and not permission.matching_members:
                permission.delete()

#------------------------------------------------------------------------------

step = MigrationStep("Remove Template.engine")

@when(step.executing)
def remove_template_engine(e):

    from woost.models import Template

    for template in Template.select():
        try:
            del template._engine
        except AttributeError:
            pass

#------------------------------------------------------------------------------

step = MigrationStep("Remove publication schemes")
migration_steps["Instrument non relational collections"].require(step)

@when(step.executing)
def remove_publication_schemes(e):

    from cocktail.persistence import datastore
    from woost.models import (
        Item,
        MemberPermission,
        Configuration
    )
    from woost.models.utils import remove_broken_type

    remove_broken_type(
        "woost.models.publicationschemes.PublicationScheme",
        existing_bases = (
            Item,
        )
    )

    remove_broken_type(
        "woost.models.publicationschemes.HierarchicalPublicationScheme",
        existing_bases = (
            Item,
        )
    )

    remove_broken_type(
        "woost.models.publicationschemes.IdPublicationScheme",
        existing_bases = (
            Item,
        )
    )

    remove_broken_type(
        "woost.models.publicationschemes.DescriptiveIdPublicationScheme",
        existing_bases = (
            Item,
        )
    )

    try:
        del Configuration.instance._publication_schemes
    except AttributeError:
        pass

    members = (
        "woost.models.configuration.Configuration.publication_schemes",
    )

    for permission in MemberPermission.select():
        if permission.matching_members:
            removed = False

            for full_member_name in members:
                try:
                    permission.matching_members.remove(full_member_name)
                except (KeyError, ValueError):
                    pass
                else:
                    removed = True

            if removed and not permission.matching_members:
                permission.delete()

#------------------------------------------------------------------------------

step = MigrationStep("Restrict access to Block.initialization")

@when(step.executing)
def restrict_block_initialization(e):
    from woost.models import Role, ReadMemberPermission
    role = Role.require_instance(qname = "woost.everybody")

    for permission in role.permissions:
        if (
            isinstance(permission, ReadMemberPermission)
            and not permission.authorized
        ):
            break
    else:
        permission = ReadMemberPermission()
        permission.authorized = False
        permission.insert()

    permission.matching_members.add(
        "woost.models.block.Block.initialization"
    )



#------------------------------------------------------------------------------

step = MigrationStep("Support per language publication in blocks")

@when(step.executing)
def remove_publication_schemes(e):

    from woost.models import Block

    for block in Block.select():
        block.per_language_publication = False
        for language in block.translations:
            block.set("translation_enabled", True, language)

#------------------------------------------------------------------------------

step = MigrationStep("Initialize last_translation_update_time")

@when(step.executing)
def initialize_last_translation_update_time(e):

    from datetime import datetime
    from woost.models import Item
    now = datetime.now()

    for item in Item.select():
        if item.__class__.translated:
            for lang in item.translations:
                last_update = item.get("last_translation_update_time", lang)
                if last_update is None:
                    item.set("last_translation_update_time", now, lang)

#------------------------------------------------------------------------------

step = MigrationStep("Rename EmailTemplate.language to .language_expression")

@when(step.executing)
def rename_email_template_language(e):
    from woost.models import EmailTemplate

    for tmpl in EmailTemplate.select():
        try:
            value = tmpl._language
        except AttributeError:
            pass
        else:
            del tmpl._language
            tmpl._language_expression = value

#------------------------------------------------------------------------------

step = MigrationStep("Ditch ContentPermission.matching_items")

@when(step.executing)
def ditch_content_permission_matching_items(e):

    from warnings import warn
    from cocktail.pkgutils import import_object
    from woost.models import Item, ContentPermission

    for permission in ContentPermission.select():
        matching_items = permission._matching_items.copy()

        type_ref = matching_items.pop("type", None)
        if type_ref is None:
            permission.content_type = Item
        else:
            permission.content_type = import_object(type_ref)

        if not matching_items:
            del permission._matching_items
        else:
            warn(
                "Permission #%d for type %s "
                "had custom filters that couldn't be migrated: %r"
                % (permission.id, type_ref, matching_items)
            )

#------------------------------------------------------------------------------

step = MigrationStep("Transform translation_enabled into enabled_translations")

@when(step.executing)
def initialize_enabled_translations(e):

    from cocktail.translations import descend_language_tree
    from woost.models import Publishable, Block

    for cls in (Publishable, Block):
        for item in cls.select():
            for lang, translation in item.translations.iteritems():
                if (
                    not item.per_language_publication
                    or getattr(translation, "_translation_enabled", False)
                ):
                    item.enabled_translations.add(lang)

                try:
                    del translation._translation_enabled
                except AttributeError:
                    pass

            for lang in list(item.enabled_translations):
                item.enabled_translations.update(
                    descend_language_tree(lang, False)
                )

#------------------------------------------------------------------------------

step = MigrationStep("Index Change.is_explicit_change")

@when(step.executing)
def index_is_explicit_change(e):
    from woost.models import Change
    Change.is_explicit_change.rebuild_index()

#------------------------------------------------------------------------------

step = MigrationStep("Introduce Publishable.access_level")

@when(step.executing)
def introduce_publishable_access_level(e):

    from warnings import warn
    from woost.models import Role, AccessLevel, Publishable, ReadPermission
    from woost.extensions.restrictedaccess import RestrictedAccessExtension

    # Fix the "Everybody can read publishable objects" permission to take
    # access levels into account.
    everybody = Role.require_instance(qname = "woost.everybody")

    for permission in everybody.permissions:
        if (
            isinstance(permission, ReadPermission)
            and permission.content_type is Publishable
            and permission.authorized
            and not permission.content_expression
        ):
            permission.content_expression = (
                "from woost.models import user_has_access_level\n"
                "items.add_filter(user_has_access_level)"
            )
            break
    else:
        warn(
            u"The migration expected to find a permission in the 'Everybody' "
            u"role that allowed all users to read publishable objects, so it "
            u"could be modified to take access levels into account. Since "
            u"that permission is not present, you will have to create it "
            u"manually, or otherwise access levels won't work. Check the "
            u"SiteInitializer.create_everybody_role() method for clues on "
            u"how the permission should work."
        )

    # Create the "Only for editors" access level
    editors = Role.get_instance(qname = "woost.editors")
    if editors is not None:
        level = AccessLevel()
        level.roles_with_access.append(editors)
        level.insert()
    else:
        warn(
            u"The migration wanted to create an 'Only for editors' access "
            u"level, but the 'Editors' role can't be found. If you want it, "
            u"you will have to create the access level on your own."
        )

    # Import access levels from the 'restrictedaccess' extension
    if RestrictedAccessExtension.instance.enabled:

        RestrictedAccessExtension.instance.enabled = False

        from woost.extensions.restrictedaccess.accessrestriction \
            import  AccessRestriction

        mapping = {}

        for restriction in AccessRestriction.select():
            level = AccessLevel()
            for lang in restriction.translations:
                level.set("title", restriction.get("title", lang), lang)
            level.insert()
            mapping[restriction] = level

        for publishable in Publishable.select():
            if publishable.access_restriction:
                publishable.access_level = \
                    mapping[publishable.access_restriction]
            try:
                del publishable.access_restriction
            except AttributeError:
                pass

        AccessRestriction.select().delete_items()

        warn(
            u"The deprecated 'restrictedaccess' extension has been disabled. "
            u"Instances of AccessRestriction have been replaced with "
            u"equivalent AccessLevel instances. You should relate the created "
            u"levels to their roles, and delete or correct the existing "
            u"permissions that were used to enforce access restrictions using "
            u"the AccessRestriction model.\n\nAlso, *remember to reboot your "
            u"site*!"
        )

    # Rebuild indexes
    Publishable.access_level.rebuild_index()

#------------------------------------------------------------------------------

step = MigrationStep("Index Publishable.parent")

@when(step.executing)
def index_publishable_parent(e):
    from woost.models import Publishable
    Publishable.parent.rebuild_index()

#------------------------------------------------------------------------------

step = MigrationStep(
    "Use sets for Website.specific_content and Publishable.websites"
)

@when(step.executing)
def use_sets_for_publishable_website_relations(e):
    from woost.models import Website, Publishable

    for website in Website.select():
        del website._specific_content

    pub_websites = {}

    for publishable in Publishable.select():
        pub_websites[publishable] = set(publishable.websites)
        del publishable._websites

    for publishable, websites in pub_websites.iteritems():
        publishable.websites = websites


#------------------------------------------------------------------------------

step = MigrationStep("Fix 'author' member in Change.item_state")

@when(step.executing)
def fix_author_in_change_item_state(e):
    from woost.models import Change
    for change in Change.select():
        if (
            change.action == "create"
            and change.target is not None
            and change.item_state is not None
            and change.item_state.get("author") is None
        ):
            change.item_state["author"] = change.target.author

#------------------------------------------------------------------------------

step = MigrationStep("Changeset indexing")

@when(step.executing)
def index_changesets(e):
    from woost.models import ChangeSet, Change
    for change in Change.select():
        ChangeSet.changes_index[change.id] = change.changeset.id
    ChangeSet.rebuild_indexes()
    Change.rebuild_indexes()

#------------------------------------------------------------------------------

step = MigrationStep("Make Item.translations not versioned")

@when(step.executing)
def make_item_translations_not_versioned(e):
    from woost.models import Change
    for change in Change.select(Change.action.equal("modify")):
        change.changed_members.discard("translations")
        change.item_state.pop("translations", None)

#------------------------------------------------------------------------------

step = MigrationStep("Make anonymous relation ends not versioned")

@when(step.executing)
def make_anonymous_relation_ends_not_versioned(e):

    from cocktail import schema
    from woost.models import Change

    keys_cache = {}

    for change in Change.select(Change.action.equal("modify")):
        model = change.target.__class__
        keys = keys_cache.get(model)
        if keys is None:
            keys = [
                member.name
                for member in model.iter_members()
                if isinstance(member, schema.RelationMember)
                and member.anonymous
            ]
            keys_cache[model] = keys

        for key in keys:
            change.changed_members.discard(key)
            change.item_state.pop(key, None)

#------------------------------------------------------------------------------

step = MigrationStep(
    "woost.extensions.googleanalytics: Create default custom definitions"
)

@when(step.executing)
def create_default_custom_definitions(e):

    from woost.extensions.googleanalytics import GoogleAnalyticsExtension
    extension = GoogleAnalyticsExtension.instance

    if extension.enabled:
        from woost.extensions.googleanalytics import installationstrings
        extension._create_default_custom_definitions()
        warn(
            "You must declare any custom dimensions/metrics that were added "
            "by your site using the declaring_tracker event. Also, be sure "
            "to replicate any definition using the administration panel of "
            "your Google Analytics property."
        )

#------------------------------------------------------------------------------

step = MigrationStep("Grant debug permission to administrators and editors")

@when(step.executing)
def grant_debug_permission(e):

    from woost.models import Role, DebugPermission

    for key in "editors", "administrators":
        role = Role.get_instance(qname = "woost." + key)
        permission = DebugPermission()
        permission.insert()
        role.permissions.append(permission)

#------------------------------------------------------------------------------

step = MigrationStep("Create generic error page")

@when(step.executing)
def create_generic_error_page(e):

    from woost.models import Configuration, CustomBlock
    from woost.models.initialization import SiteInitializer

    config = Configuration.instance

    if config.generic_error_page is None:
        init = SiteInitializer()
        init.languages = config.languages
        config.generic_error_page = init.create_generic_error_page()
    else:
        tb_block = CustomBlock()
        tb_block.view_class = "woost.views.Traceback"
        tb_block.insert()
        config.generic_error_page.blocks.append(tb_block)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from collections import Counter
from cocktail.translations import translations
from woost.models import Item, Configuration
from woost.controllers.notifications import Notification
from .state import TranslationWorkflowState

def get_models_included_in_translation_workflow(root_model = Item):
    return tuple(
        model
        for model in root_model.schema_tree()
        if model_is_included_in_translation_workflow(model)
    )

def model_is_included_in_translation_workflow(model):
    return (
        model.included_in_translation_workflow
        and any(
            member_is_included_in_translation_workflow(member)
            for member in model.iter_members()
        )
    )

def member_is_included_in_translation_workflow(member):
    return member.translated and member.included_in_translation_workflow

def object_is_included_in_translation_workflow(obj):
    return (
        obj.included_in_translation_workflow
        and isinstance(obj, get_models_included_in_translation_workflow())
    )

def iter_changeset_translation_requests(
    changeset,
    include_silenced = False,
    include_unchanged = False
):
    from .request import TranslationWorkflowRequest

    # Yield created or modified requests
    for change in changeset.changes.itervalues():
        if isinstance(change.target, TranslationWorkflowRequest):
            yield change.target

    # Yield silenced requests that would have been affected by the changeset
    if include_silenced or include_unchanged:
        silenced_state = TranslationWorkflowState.get_instance(
            qname = "woost.extensions.translationworkflow.states.silenced"
        )

        for item_id, change in changeset.changes.iteritems():
            if (
                change.action == "modify"
                and change.target
                and change.target.is_inserted
                and object_is_included_in_translation_workflow(change.target)
            ):
                source_languages = set()
                for member, language in change.diff():
                    if member_is_included_in_translation_workflow(member):
                        source_languages.add(language)

                language_paths = \
                    Configuration.instance.translation_workflow_paths

                for source_language in source_languages:
                    target_languages = language_paths.get(source_language)
                    if target_languages:
                        for target_language in target_languages:
                            request = change.target.get_translation_request(
                                source_language,
                                target_language
                            )
                            if request is not None:
                                if (
                                    (
                                        silenced_state is not None
                                        and request.state is silenced_state
                                    )
                                    or include_unchanged
                                ):
                                    yield request

def iter_changeset_translation_request_changes(
    changeset,
    include_silenced = False,
    include_unchanged = False,
    only_changed_states = False
):
    silenced_state = TranslationWorkflowState.get_instance(
        qname = "woost.extensions.translationworkflow.states.silenced"
    )

    for request in iter_changeset_translation_requests(changeset):
        request_change = changeset.changes.get(request.id)
        if request_change is None:
            if (
                silenced_state is not None
                and request.state is silenced_state
            ):
                if include_unchanged:
                    yield request, "silenced"
            elif include_unchanged:
                yield request, "unchanged"
        elif request_change.action == "create":
            yield request, "created",
        elif request_change.action == "modify":
            if (
                not only_changed_states
                or "state" in request_change.changed_members
            ):
                yield request, "invalidated"

def notify_translation_request_changes(changeset):

    silenced_count = 0
    state_count = Counter()

    for request, request_change \
    in iter_changeset_translation_request_changes(
        changeset,
        only_changed_states = True
    ):
        if request_change == "silenced":
            silenced_count += 1
        else:
            state_count[request.state] += 1

    for state, count in state_count.iteritems():
        Notification(
            translations(
                "woost.extensions.translationworkflow.transitions_notice",
                new_state = state,
                count = count
            ),
            "success"
        ).emit()

    if silenced_count:
        Notification(
            translations(
                "woost.extensions.translationworkflow.silence_notice",
                count = silenced_count
            )
        ).emit()




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import os
from glob import glob
from threading import Lock, Event
from cocktail.styled import styled
from cocktail.iteration import first
from cocktail.events import when
from cocktail.persistence import datastore
from woost import app
from woost.models import Configuration, File, User, ReadPermission
from woost.extensions.audio.audioencoder import AudioEncoder

_cache_lock = Lock()
_cache_ongoing_requests = {}

debug = False

def get_audio_uri(file, encoding):

    if isinstance(encoding, basestring):

        # Format requested by file extension
        if encoding.startswith("."):

            # The source file already matches the requested file extension,
            # use it as is
            if file.file_extension == encoding:
                return file.get_uri()

            # Otherwise, use the first encoder that matches the given file
            # extension
            encoder = first(
                enc
                for enc in Configuration.instance.audio_encoders
                if enc.extension == encoding
            )

        # Format requested by the user defined encoder ID (ie. mp3-128)
        else:
            encoder = AudioEncoder.require_instance(identifier = encoding)
    else:
        encoder = encoding

    if encoder is None:
        return None
    else:
        return "/audio/%d/%s%s" % (
            file.id,
            encoder.identifier,
            encoder.extension
        )

def request_encoding(file, encoder):

    if isinstance(encoder, basestring):
        encoder = AudioEncoder.require_instance(identifier = encoder)

    dest = get_audio_cache_dest(file, encoder)

    # The file hasn't been generated yet
    if not os.path.exists(dest):

        request = None
        request_key = (file.id, encoder.identifier)

        with _cache_lock:
            request = _cache_ongoing_requests.get(request_key)

            # Another thread is already requesting this file encoding: wait
            # for it to finish
            if request is not None:
                new_request = False
            # No other thread is generating the requested file/encoding
            # combination at present: generate it
            else:
                request = Event()
                _cache_ongoing_requests[request_key] = request
                new_request = True

                # Make sure the cache folder for the indicated file exists
                folder = app.path("audio-cache", str(file.id))

                if not os.path.exists(folder):
                    os.mkdir(folder)

        if new_request:
            try:
                encoder.encode(file, dest)
                with _cache_lock:
                    _create_symlink(file, dest)
                    request.set()
            finally:
                try:
                    del _cache_ongoing_requests[request_key]
                except KeyError:
                    pass
        else:
            request.wait()

    return dest

def get_audio_cache_dest(file, encoder):
    return app.path(
        "audio-cache",
        str(file.id),
        encoder.identifier + encoder.extension
    )

def _create_symlink(file, dest):
    """Create a symbolic link for static publication of an encoded file."""

    if not hasattr(os, "symlink"):
        return

    # Only create static publication links for public content
    anonymous = User.require_instance(qname = "woost.anonymous_user")
    if not anonymous.has_permission(ReadPermission, target = file):
        return

    static_folder = app.path("static", "audio", str(file.id))
    static_link = os.path.join(static_folder, os.path.basename(dest))

    if not os.path.lexists(static_link):

        # Make sure the containing folder exists
        if not os.path.exists(static_folder):
            os.mkdir(static_folder)

        os.symlink(dest, static_link)

def _remove_dir_contents(path, pattern = None):

    # os.path.lexists is not supported on windows
    from cocktail.styled import styled
    exists = getattr(os.path, "lexists", os.path.exists)

    if exists(path):

        if pattern:
            items = glob(os.path.join(path, pattern))
        else:
            items = os.listdir(path)

        for item in items:
            item_path = os.path.join(path, item)
            if debug:
                print styled(" " * 4 + item_path, "red")
            try:
                if os.path.isdir(item_path):
                    rmtree(item_path)
                else:
                    os.remove(item_path)
            except OSError, ex:
                if debug:
                    print styled(ex, "red")

def clear_audio_cache(item = None, encoder_id = None):

    if debug:
        from cocktail.styled import styled
        print styled("Clearing audio cache", "red"),

        if item:
            print styled("Item:", "light_gray"),
            print styled(item, "red", style = "bold"),

        if encoder_id:
            print styled("Encoder:", "light_gray"),
            print styled(encoder_id, "red", style = "bold"),

        print

    # Remove the full cache
    if item is None and encoder_id is None:
        _remove_dir_contents(app.path("audio-cache"))
        _remove_dir_contents(app.path("static", "audio"))

    # Selective drop: per item and/or encoder
    else:
        paths = []

        if item is not None:
            paths.append(app.path("audio-cache", str(item.id)))
            paths.append(app.path("static", "audio", str(item.id)))
        else:
            for base in (
                app.path("audio-cache"),
                app.path("static", "audio")
            ):
                for item in os.listdir(base):
                    path = os.path.join(base, item)
                    if os.path.isdir(path):
                        paths.append(path)

        if encoder_id is None:
            pattern = None
        else:
            pattern = encoder_id + ".*"

        for path in paths:
            _remove_dir_contents(path, pattern)

@when(File.changed)
@when(File.deleted)
def _clear_audio_cache_after_commit(event):
    item = event.source
    if item.is_inserted:
        clear_audio_cache_after_commit(item = item)

def clear_audio_cache_after_commit(item = None, encoder = None):

    if encoder:
        encoder_id = encoder.identifier
    else:
        encoder_id = None

    key = "woost.extensions.audio.request.clear_audio_cache(item=%s,encoder=%s)" % (
        item.id if item else "*",
        encoder_id or "*"
    )

    datastore.unique_after_commit_hook(
        key,
        _clear_audio_cache_after_commit_callback,
        item,
        encoder_id
    )

def _clear_audio_cache_after_commit_callback(
    commit_successful,
    item,
    encoder_id
):
    if commit_successful:
        clear_audio_cache(item, encoder_id)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cStringIO import StringIO
import mimetypes
import cherrypy
from cocktail.translations import translations
from cocktail import schema
from cocktail.schema.io import export_msexcel
from cocktail.controllers import request_property, get_parameter
from woost.controllers.backoffice.basebackofficecontroller \
    import BaseBackOfficeController
from .formblock import FormBlock


class ExportFormDataController(BaseBackOfficeController):

    @request_property
    def form(self):
        return get_parameter(
            schema.Reference("form", type = FormBlock, required = True),
            errors = "raise"
        )

    def __call__(self, *args, **kwargs):

        form = self.form
        form_model = form.field_set.create_form_model()

        cherrypy.response.headers['Content-Type'] = \
            mimetypes.types_map.get(".xls")

        cherrypy.response.headers["Content-Disposition"] = \
            'attachment; filename="%s.xls"' % translations(form)

        buffer = StringIO()
        export_msexcel(
            form.submitted_data,
            buffer,
            form_model,
            form_model.ordered_members()
        )
        return buffer.getvalue()




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import (
    Document,
    Slot,
    Template,
    Controller,
    CustomBlock,
    HTMLBlock,
    block_type_groups
)
from .newslettercontent import NewsletterContent
from .newsletterbox import NewsletterBox
from .newsletterlisting import NewsletterListing


class Newsletter(Document):

    default_template = schema.DynamicDefault(
        lambda: Template.get_instance(
            qname = "woost.extensions.newsletters.newsletter_template"
        )
    )

    default_controller = schema.DynamicDefault(
        lambda: Controller.get_instance(
            qname = "woost.extensions.newsletters.newsletter_controller"
        )
    )

    members_order = [
        "blocks"
    ]

    blocks = Slot()

    allowed_block_types = (
        NewsletterBox,
        NewsletterContent,
        NewsletterListing,
        CustomBlock,
        HTMLBlock
    )

    def allows_block_type(self, block_type):
        return issubclass(block_type, self.allowed_block_types)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Item, Publishable


class FormAgreement(Item):

    type_group = "custom_forms"
    show_element_in_listing = False

    members_order = [
        "forms",
        "title",
        "document"
    ]

    forms = schema.Collection(
        items = "woost.extensions.forms.formblock.FormBlock",
        bidirectional = True,
        editable = schema.NOT_EDITABLE
    )

    title = schema.String(
        required = True,
        descriptive = True,
        translated = True,
        unique = True,
        indexed = True,
        spellcheck = True,
        normalized_index = True
    )

    document = schema.Reference(
        type = Publishable,
        related_end = schema.Collection()
    )




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.translations import translations
from cocktail.html import templates
from woost.models import Extension


translations.define("IdentityExtension",
    ca = u"Identitat",
    es = u"Identidad",
    en = u"Identity"
)

translations.define("IdentityExtension-plural",
    ca = u"Identitat",
    es = u"Identidad",
    en = u"Identity"
)


class IdentityExtension(Extension):

    def __init__(self, **values):
        Extension.__init__(self, **values)
        self.extension_author = u"Whads/Accent SL"
        self.set("description",
            u"""
                Integració de serveis d'autenticació externs (Facebook,
                Google, etc).
            """,
            "ca"
        )
        self.set("description",
            u"""
                Integración de servicios de autenticación externos (Facebook,
                Google, etc).
                """,
            "es"
        )
        self.set("description",
            u"""
                Integration with external authentication providers (Facebook,
                Google, etc).
            """,
            "en"
        )

    def _load(self):

        from woost.extensions.identity import (
            strings,
            configuration,
            identityprovider,
            facebook,
            google
        )

        templates.get_class("woost.extensions.identity.LoginBlockViewOverlay")
        templates.get_class("woost.extensions.identity.ItemCardOverlay")




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Item


class AccessRestriction(Item):

    instantiable = True
    type_group = "users"

    members_order = [
        "title",
        "restricted_content"
    ]

    title = schema.String(
        translated = True,
        required = True,
        unique = True,
        indexed = True,
        spellcheck = True,
        descriptive = True
    )

    restricted_content = schema.Collection(
        items = "woost.models.Publishable",
        bidirectional = True,
        editable = schema.NOT_EDITABLE
    )




#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			October 2008
"""
import cherrypy
from cocktail.translations import get_language, translations
from cocktail.modeling import cached_getter
from cocktail.controllers import get_state
from woost import app
from woost.controllers.publishablecontroller import PublishableController


class DocumentController(PublishableController):
    """A controller that serves rendered pages."""

    def __call__(self, **kwargs):

        # Document specified redirection
        document = app.publishable

        if document.redirection_mode:

            redirection_target = document.find_redirection_target()

            if redirection_target is None:
                raise cherrypy.NotFound()

            if redirection_target.is_internal_content():
                parameters = get_state()
            else:
                parameters = None

            uri = redirection_target.get_uri(parameters = parameters)
            method = document.redirection_method

            if method == "perm":
                raise cherrypy.HTTPRedirect(uri, 301)
            elif method == "client":
                return """
                    <!DOCTYPE html>
                    <html lang="%s">
                        <head>
                            <meta http-equiv="Content-Type" content="text/html;charset=utf-8"/>
                            <meta http-equiv="refresh" content="1; url=%s"/>
                            <title>%s</title>
                        </head>
                        <body>
                            %s
                            <script type="text/javascript">
                                location.href = document.getElementById("woost-client-redirection").href;
                            </script>
                        </body>
                    </html>
                """ % (
                    get_language(),
                    uri,
                    translations("woost.controllers.DocumentController.redirection_title"),
                    translations(
                        "woost.controllers.DocumentController.redirection_explanation",
                        uri = uri
                    )
                )
            else:
                raise cherrypy.HTTPRedirect(uri)

        # No redirection, serve the document normally
        return PublishableController.__call__(self)

    @cached_getter
    def page_template(self):
        template = app.publishable.template

        if template is None:
            raise cherrypy.NotFound()

        return template

    @cached_getter
    def view_class(self):
        return self.page_template.identifier




#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			February 2009
"""
from cocktail.html import templates
from cocktail.html.uigeneration import UIGenerator
from woost import app
from woost.models import ReadPermission
from woost.views.uigeneration import backoffice_display

List = templates.get_class("cocktail.html.List")


class ContentList(List, UIGenerator):

    referer = None
    base_ui_generators = [backoffice_display]

    def __init__(self, *args, **kwargs):
        List.__init__(self, *args, **kwargs)
        UIGenerator.__init__(self)

    def _fill_entries(self):

        user = app.user
        items = self.items

        if items is not None:
            items = [
                item
                for item in self.items
                if user.has_permission(ReadPermission, target = item)
            ]

        if items:
            List._fill_entries(self)
        else:
            self.tag = "div"
            self.append(u"-")

    def create_entry_content(self, item):

        if self.member is not None and self.member.items is not None:
            display = self.create_member_display(
                None,
                self.member.items,
                item,
                referer = self.referer or self.data
            )
        else:
            display = self.create_object_display(
                item,
                referer = self.referer or self.data
            )

        if (
            self.referer
            and self.member
            and getattr(display, "item_label", None)
        ):
            display.item_label["data-woost-relativedrop"] = "%d.%s.%d" % (
                self.referer.id,
                self.member.name,
                item.id
            )

        return display




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.modeling import cached_getter
from woost import app
from woost.models import User
from .translationagency import TranslationAgency


class TranslationWorkflowTransitionSetup(object):

    def __init__(self, transition, requests):
        self.transition = transition
        self.requests = requests

    @cached_getter
    def multiple_choices(self):
        return True

    @cached_getter
    def transition_schema(self):
        return None

    @cached_getter
    def translation_paths(self):
        return set(
            (request.source_language, request.target_language)
            for request in self.requests
        )

    @cached_getter
    def eligible_translators(self):

        user = app.user
        agency = user.managed_translation_agency

        # Coordinators can assign to freelance translators and to translation
        # agencies
        if agency is None:
            all_translators = (
                list(TranslationAgency.select())
                + list(User.select({
                    "enabled": True,
                    "translation_agency": None
                }))
            )
        # Translation agency managers can assign to their agency's translators
        else:
            all_translators = agency.translators

        # Filter translators / agencies by language proficiencies
        return dict(
            (translation_path, [
                translator
                for translator in all_translators
                if translation_path in translator.translation_proficiencies
            ])
            for translation_path in self.translation_paths
        )




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.translations import translations
from cocktail.html import Element
from woost.models import get_current_user, ReadPermission


class Link(Element):

    NO_CHECK = 0
    REQUIRE_PUBLISHED = 1
    REQUIRE_READABLE = 2
    REQUIRE_ACCESSIBLE = REQUIRE_PUBLISHED | REQUIRE_READABLE

    NO_EFFECT = 0
    HIDE = 1
    BECOME_TRANSPARENT = 2
    BECOME_SPAN = 3

    tag = "a"
    value = None
    language = None
    host = None
    parameters = None
    path = None
    content_check = REQUIRE_ACCESSIBLE
    inactive_behavior = HIDE

    def _ready(self):

        active = (
            self.value is not None
            and not (
                self.content_check & self.REQUIRE_PUBLISHED
                and not self.value.is_published(self.language)
            )
            and not (
                self.content_check & self.REQUIRE_READABLE
                and not get_current_user().has_permission(
                    ReadPermission,
                    target = self.value
                )
            )
        )

        if not active:
            self._deactivate_link()

        Element._ready(self)

        if active:
            uri = self.get_uri()
            if uri:
                self["href"] = uri
            else:
                self._deactivate_link()

    def _deactivate_link(self):
        if self.inactive_behavior == self.HIDE:
            self.visible = False
        elif self.inactive_behavior == self.BECOME_TRANSPARENT:
            self.tag = None
        elif self.inactive_behavior == self.BECOME_SPAN:
            self.tag = "span"
        elif self.inactive_behavior != self.NO_EFFECT:
            raise ValueError(
                "Unsupported value for Link.inactive_behavior (%r)"
                % self.inactive_behavior
            )

    def get_uri(self):
        if not self.value:
            return None
        else:
            return self.value.get_uri(
                language = self.language,
                host = self.host,
                path = self.path,
                parameters = self.parameters
            )




#-*- coding: utf-8 -*-
"""

@author:		Jordi Fernández
@contact:		jordi.fernandez@whads.com
@organization:	Whads/Accent SL
@since:			March 2010
"""
import os
from shutil import rmtree
from subprocess import Popen, PIPE
from cocktail import schema
from cocktail.translations import language_context
from cocktail.controllers import context as controller_context
from cocktail.modeling import abstractmethod, getter
from cocktail.translations import translations
from woost import app
from woost.models import Item


class StaticSiteSnapShoter(Item):
    """A class tasked with creating a static snapshot of a site's content to
    a concrete location.

    This is mostly an abstract class, meant to be extended by subclasses. Each
    subclass should implement the static snapshot creation.
    """
    instantiable = False
    visible_from_root = False
    integral = True

    members_order = [
        "setup_expression",
        "postprocessing_expression"
    ]

    setup_expression = schema.CodeBlock(
        language = "python"
    )

    postprocessing_expression = schema.CodeBlock(
        language = "python"
    )

    def setup(self, context):
        """Prepares the exporter for an export process.

        The object of this method is to allow exporters to perform any
        initialization they require before writing files to their destination.

        @param context: A dictionary where the exporter can place any
            contextual information it many need throgout the export process. It
            will be made available to all L{write_file} calls.
        @type context: dict
        """
        setup_expression = self.setup_expression
        if setup_expression:
            setup_expression = compile(
                setup_expression,
                "%s #%d.setup_expression" % (self.__class__.__name__, self.id),
                "exec"
            )
            exec setup_expression in context

    def cleanup(self, context):
        """Frees resources after an export operation.

        This method is guaranteed to be called after the export operation is
        over, even if an exception was raised during its execution.

        @param context: The dictionary used by the exporter during the export
            process to maintain its contextual information.
        @type context: dict
        """

    def snapshot(self, items, context = None):
        """ Generates the snapshot of a site's content

        @param items: The list of items which the exportation will start.
        @type items: L{Publishable}

        @param context: A dictionary used to share any contextual information
            with the snapshoter.
        @type context: dict
        """
        if context is None:
            context = {}

        self.setup(context)

        try:
            return self._snapshot(items, context)
        finally:
            self.cleanup(context)

    @abstractmethod
    def _snapshot(self, items, context):
        """ Generates the snapshot of a site's content

        @param items: The list of items which the exportation will start.
        @type items: L{Publishable}

        @param context: A dictionary used to share any contextual information
            with the snapshoter.
        @type context: dict
        """

    def _yield_directory(self, directory, context):

        postprocessing = self.postprocessing_expression
        if postprocessing:
            postprocessing = compile(
                postprocessing,
                "%s #%d.postprocessing_expression" % (
                    self.__class__.__name__,
                    self.id
                ),
                "exec"
            )

        for root, dirs, files in os.walk(directory):
            for file in files:
                file_path = os.path.join(root, file)

                if isinstance(file_path, str):
                    file_path = file_path.decode("utf-8")

                relative_path = os.path.relpath(file_path, directory)

                if postprocessing is not None:
                    exec_context = context.copy()
                    exec_context["file_path"] = file_path
                    exec_context["relative_path"] = relative_path
                    exec postprocessing in exec_context

                yield (file_path, relative_path)


class WgetSnapShoter(StaticSiteSnapShoter):
    """ A class that creates a static snapshot of a site's content using wget """
    instantiable = True

    file_names_mode = schema.String(
        required = True,
        default = "unix",
        text_search = False,
        enumeration = frozenset(("unix", "windows")),
        translate_value = lambda value, **kwargs:
            u"" if not value else translations(
                "woost.extensions.staticsite.staticsitesnapshoter.WgetSnapShoter.file_names_mode " + value,
                **kwargs
            )
    )

    @getter
    def snapshot_path(self):
        return app.path("snapshots", str(self.id))

    def _get_cmd(self, context):
        cmd = "wget "

        if context.get("follow_links"):
            cmd += " --mirror"

        cmd += (
            " --header='X-Woost-StaticSiteSnapshoter: " + self.__class__.__name__ + "'"
            " --page-requisites"
            " --html-extension"
            " --convert-links"
            " --no-host-directories"
            " --no-check-certificate"
            " --trust-server-names"
            " --directory-prefix=%s"
            " --restrict-file-names=%s %s"
        )

        return cmd

    def _snapshot(self, items, context = {}):

        snapshot_path = self.snapshot_path

        if os.path.exists(snapshot_path):
            rmtree(snapshot_path)

        os.mkdir(snapshot_path)

        cmd = self._get_cmd(context)
        uris = set()

        for item in items:
            if isinstance(item, basestring):
                uris.add(item)
            else:
                if item.per_language_publication:
                    for language in item.enabled_translations:
                        with language_context(language):
                            uris.add(self._get_uri(item, context))
                else:
                    uris.add(self._get_uri(item, context))

        cmd = cmd % (
            snapshot_path,
            self.file_names_mode,
            u" ".join(uris)
        )

        p = Popen(cmd, shell=True, stdout=PIPE)
        p.communicate()

        for item in self._yield_directory(self.snapshot_path, context):
            yield item

    def _get_uri(self, item, context):
        uri = item.get_uri(host = "!")
        if isinstance(uri, unicode):
            uri = uri.encode("utf-8")
        return uri




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.pkgutils import resolve
from cocktail.translations import require_language
from cocktail.html.autocomplete import Autocomplete as BaseAutocomplete


class Autocomplete(BaseAutocomplete):

    ajax_search = True
    ajax_search_threshold = 100
    show_icons = True
    show_types = True
    icon_factory = "icon16"

    def _build(self):
        BaseAutocomplete._build(self)
        self.add_resource("/resources/scripts/Autocomplete.js")

    def _ready(self):
        BaseAutocomplete._ready(self)
        self["data-woost-autocomplete-show-types"] = \
            "true" if self.show_types else "false"
        self["data-woost-autocomplete-show-icons"] = \
            "true" if self.show_icons else "false"
        self["data-woost-autocomplete-icon-factory"] = self.icon_factory
        if self.show_types:
            self.add_resource(
                "/cms_metadata?language=%s&format=javascript"
                % require_language(),
                mime_type = "text/javascript"
            )

    def get_default_ajax_url(self):
        return (
            "/autocomplete/%s/QUERY?lang=%s"
            % (
                self.member.original_member.get_qualified_name(
                    include_ns = True
                ),
                require_language()
            )
        )

    def create_autocomplete_source(self):
        autocomplete_class = resolve(self.member.autocomplete_class)
        return autocomplete_class(self.member)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from urllib import urlopen
from httplib import HTTPConnection
from time import mktime, strptime
from cStringIO import StringIO
from PIL import Image
from cocktail import schema
from woost.models.uri import URI
from woost.models.rendering.renderer import Renderer


class ImageURIRenderer(Renderer):

    instantiable = True

    def can_render(self, item, **parameters):
        return isinstance(item, URI) and item.resource_type == "image"

    def get_item_uri(self, item):
        return item.uri

    def render(self, item, **parameters):

        # Open the remote resource
        uri = self.get_item_uri(item)
        http_resource = urlopen(uri)

        # Wrap image data in a buffer
        # (the object returned by urlopen() doesn't support seek(), which is
        # required by PIL)
        buffer = StringIO()
        buffer.write(http_resource.read())
        buffer.seek(0)

        return Image.open(buffer)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Role
from .state import TranslationWorkflowState
from .transition import TranslationWorkflowTransition

Role.add_member(
    schema.Collection("translation_workflow_relevant_states",
        items = schema.Reference(
            type = TranslationWorkflowState
        ),
        related_end = schema.Collection()
    ),
    append = True
)

Role.add_member(
    schema.Reference("translation_workflow_default_state",
        type = TranslationWorkflowState,
        related_end = schema.Collection(),
        enumeration = Role.translation_workflow_relevant_states,
        listed_by_default = False
    ),
    append = True
)

Role.add_member(
    schema.Reference("translation_workflow_default_transition",
        type = TranslationWorkflowTransition,
        related_end = schema.Collection(),
        listed_by_default = False
    ),
    append = True
)

Role.add_member(
    schema.CodeBlock("translation_workflow_default_transition_condition",
        language = "python",
        listed_by_default = False
    ),
    append = True
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import collections
import os
import shutil
import re
from collections import Sequence, Set
from difflib import SequenceMatcher
from ZODB.broken import Broken
from cocktail.styled import styled
from cocktail.modeling import ListWrapper, SetWrapper
from cocktail import schema
from cocktail.translations import get_language, translations
from cocktail import schema
from cocktail.schema.expressions import Self
from cocktail.persistence import (
    datastore,
    Query,
    PersistentObject,
    PersistentClass,
    reset_incremental_id,
    incremental_id
)
from cocktail.persistence.utils import (
    is_broken,
    remove_broken_type as cocktail_remove_broken_type
)
from woost import app
from .item import Item
from .configuration import Configuration
from .permission import ContentPermission
from .changesets import Change
from .role import Role
from .file import File
from . import staticpublication
from . import objectio


def remove_broken_type(
    full_name,
    existing_bases = (),
    relations = (),
    excluded_relations = None,
    languages = None
):
    if languages is None:
        languages = Configuration.instance.languages

    if excluded_relations is None:
        excluded_relations = (Change.target,)

    cocktail_remove_broken_type(
        full_name,
        existing_bases = existing_bases,
        relations = relations,
        excluded_relations = excluded_relations,
        languages = languages
    )

    for role in Role.select():
        for cls in list(role.hidden_content_types):
            if (
                issubclass(cls, Broken)
                and cls.__module__ + "." + cls.__name__ == full_name
            ):
                role.hidden_content_types.remove(cls)

    for permission in ContentPermission.select():
        content_type = permission.content_type
        if content_type.__module__ + "." + content_type.__name__ == full_name:
            permission.delete()

def delete_history():

    for item in Item.select():
        if not is_broken(item):
            try:
                del item._changes
            except AttributeError:
                pass

    for key in list(datastore.root):
        if key.startswith("woost.models.changesets."):
            del datastore.root[key]

def restore_deleted_item(obj):
    if obj.insert():
        for member in obj.__class__.iter_members():
            if (
                isinstance(member, schema.Collection)
                and obj.get(member) is None
            ):
                obj.set(member, member.produce_default(instance = obj))

def iter_text(objects = None):

    if objects is None:
        objects = Item.select()

    for obj in objects:
        for member in obj.__class__.iter_members():
            if isinstance(member, schema.String):
                if member.translated:
                    languages = obj.translations.keys()
                else:
                    languages = (None,)

                for language in languages:
                    value = obj.get(member, language)
                    if value:
                        yield obj, member, language, value

def grep(expr, objects = None):

    if isinstance(expr, basestring):
        expr = re.compile(expr)

    for obj, member, language, value in iter_text(objects):
        matches = list(expr.finditer(value))
        if matches:
            yield obj, member, language, value, matches

def hl(expr, objects = None):

    for obj, member, language, value, matches in grep(expr, objects):
        print styled("-" * 100, style = "bold")
        print styled(repr(obj), style = "bold"),
        print styled(member.name, "slate_blue"),

        if language:
            print styled(language, "pink")
        else:
            print

        hl_value = value
        offset = 0

        for match in matches:
            start, end = match.span()
            original_chunk = value[start:end]
            hl_chunk = styled(original_chunk, "magenta")
            hl_value = (
                hl_value[:start + offset]
                + hl_chunk
                + hl_value[end + offset:]
            )
            offset += len(hl_chunk) - len(original_chunk)

        print hl_value

def replace(expr, replacement, objects = None, mode = "apply"):

    if isinstance(expr, basestring):
        expr = re.compile(expr)

    if mode == "apply":
        apply = True
        show = False
    elif mode == "verbose":
        apply = True
        show = True
    elif mode == "show":
        apply = False
        show = True
    elif mode == "confirm":
        apply = None
        show = True

    for obj, member, language, value in iter_text(objects):
        modified_value = expr.sub(replacement, value)

        if value == modified_value:
            continue

        if show:
            print styled("-" * 100, style = "bold")
            print styled(translations(obj), style = "bold"),
            print styled(member.name, "slate_blue"),

            if language:
                print styled(language, "pink")
            else:
                print

            diff = SequenceMatcher(a = value, b = modified_value)
            chunks = []

            for op, start_a, end_a, start_b, end_b in diff.get_opcodes():
                orig_chunk = value[start_a:end_a]
                new_chunk = modified_value[start_b:end_b]

                if op == "equal":
                    chunks.append(orig_chunk)
                else:
                    if op in ("replace", "delete"):
                        chunks.append(styled(orig_chunk, "white", "red"))

                    if op in ("replace", "insert"):
                        chunks.append(styled(new_chunk, "white", "green"))

            print "".join(chunks)

        if apply is None:
            answer = None
            while answer not in list("yn"):
                answer = raw_input("Replace (y/n)? ")
            if answer == "n":
                continue
        elif not apply:
            continue

        obj.set(member, modified_value, language)

def search(
    query_text,
    objects = None,
    languages = None,
    match_mode = "pattern",
    stemming = None,
    **search_kwargs
):
    if objects is None:
        query = Item.select()
    elif isinstance(objects, (schema.SchemaClass, Query)):
        query = objects.select()
    elif isinstance(objects, collections.Iterable):
        query = Item.select()
        query.base_collection = objects

    if languages is None:
        languages = (None, get_language())

    query.add_filter(
        Self.search(
            query_text,
            languages = languages,
            match_mode = match_mode,
            stemming = stemming,
            **search_kwargs
        )
    )

    if stemming is None:
        stemming = query.type.stemming

    highlighter = schema.SearchHighlighter(
        query_text,
        languages,
        lambda text: styled(text, "magenta"),
        match_mode = match_mode,
        stemming = stemming
    )

    for result in query:
        print styled(result, "slate_blue")
        highlights = highlighter.highlight(result)
        if highlights:
            print highlights

def any_translation(obj, language_chain = None, **kwargs):

    if language_chain is None:
        user = app.user
        language_chain = (
            user
            and user.backoffice_language_chain
            or Configuration.instance.backoffice_language_chain
        )

    if not language_chain:
        return translations(obj, **kwargs)
    else:
        if isinstance(obj, schema.SchemaObject):
            chain_kwargs = kwargs.copy()
            chain_kwargs.setdefault("discard_generic_translation", True)
        else:
            chain_kwargs = kwargs

        for language in language_chain:
            label = translations(
                obj,
                language = language,
                **chain_kwargs
            )
            if label:
                return label
        else:
            return translations(
                obj,
                language = language_chain[0],
                **kwargs
            )

def rebase_id(base_id, verbose = False):

    # Remove all static publication links
    if verbose:
        print "Removing static publication links"

    for file in File.select():
        staticpublication.remove_links(file)

    reset_incremental_id(base_id)
    id_map = {}

    # Change ids in the database
    for root_model in PersistentObject.derived_schemas(recursive = False):
        if root_model.indexed:

            if verbose:
                print "Rebasing", root_model.__name__

            # Rebuild the object map
            pairs = list(root_model.index.items())
            root_model.index = root_model.index.__class__()
            for old_id, obj in pairs:
                id_map[old_id] = new_id = incremental_id()
                obj._id = new_id
                if verbose:
                    print "\t", old_id, ">", new_id
                root_model.index.add(new_id, obj)

            # Rebuild the .keys set for all models
            for model in root_model.schema_tree():
                if verbose:
                    print "\t", "Rebuilding .keys for", model.__name__
                keys = list(model.keys)
                model.keys.clear()
                for old_id in keys:
                    model.keys.insert(id_map[old_id])

            # Rebuild indexes
            root_model.rebuild_indexes(
                recursive = True,
                verbose = verbose
            )
            root_model.rebuild_full_text_indexes(
                recursive = True,
                verbose = verbose
            )

    if verbose:
        print "Committing transaction"

    datastore.commit()

    # Rename uploaded files
    upload_path = app.path("upload")
    uploads = []
    for file_name in os.listdir(upload_path):
        file_path = os.path.join(upload_path, file_name)
        if os.path.isfile(file_path):
            uploads.append(int(file_name))

    uploads.sort()
    uploads.reverse()

    for old_upload_id in uploads:
        new_upload_id = id_map[old_upload_id]
        old_path = os.path.join(upload_path, str(old_upload_id))
        new_path = os.path.join(upload_path, str(new_upload_id))
        if verbose:
            print "Moving upload %s to %s" % (old_path, new_path)
        shutil.move(old_path, new_path)

    # Recreate static publication links
    if verbose:
        print "Creating static publication links"

    for file in File.select():
        staticpublication.create_links(file)

def show_translation_coverage(
    objects = None,
    languages = None,
    styles = (
        (100, {'foreground': 'bright_green'}),
        (90,  {'foreground': 'yellow'}),
        (66,  {'foreground': 'brown'}),
        (33,  {'foreground': 'pink'}),
        (0,   {'foreground': 'magenta'})
    )
):
    count = collections.Counter()

    if objects is None:
        translated_models = {}

        def model_is_translated(model):
            try:
                return translated_models[model]
            except KeyError:
                for member in model.iter_members():
                    if (
                        member is not Item.last_translation_update_time
                        and member.translated
                    ):
                        translated = True
                        break
                    else:
                        translated = False
                translated_models[model] = translated
                return translated

        objects = (
            item
            for item in Item.select()
            if model_is_translated(item.__class__)
        )
    elif isinstance(objects, PersistentClass):
        objects = objects.select()

    if languages is None:
        languages = Configuration.instance.languages

    for lang in languages:
        count[lang] = 0

    for obj in objects:
        for lang in languages:
            if lang in obj.translations:
                count[lang] += 1

    results = count.most_common()
    total = results[0][1]
    table = []
    longest_label_length = 0

    for lang, n in results:
        label = translations("locale", locale = lang)
        longest_label_length = max(longest_label_length, len(label))
        percent = float(n) / total * 100
        table.append((label, percent))

    for label, percent in table:

        line = (
            (label + ":").ljust(longest_label_length + 1)
            + " "
            + ("%.2f%%" % percent).rjust(7)
        )

        if styles:
            for threshold, style in styles:
                if percent >= threshold:
                    line = styled(line, **style)
                    break

        print line

def export_json(
    obj,
    exporter = None,
    json_encoder_settings = None,
    **kwargs
):
    kwargs.setdefault("verbose", True)

    exporter = exporter or objectio.ObjectExporter()
    for key, value in kwargs.iteritems():
        setattr(exporter, key, value)

    if json_encoder_settings is None:
        json_encoder_settings = {"indent": 2, "sort_keys": True}

    if isinstance(obj, (Sequence, Set, ListWrapper, SetWrapper)):
        exporter.add_all(obj)
    elif isinstance(obj, PersistentClass):
        exporter.add_all(obj.select())
    else:
        exporter.add(obj)

    json = exporter.dumps(**json_encoder_settings)
    return json

def import_json(json, importer = None, **kwargs):

    kwargs.setdefault(
        "unknown_member_policy",
        objectio.UnknownMemberPolicy.report
    )

    kwargs.setdefault(
        "missing_object_policy",
        objectio.MissingObjectPolicy.report
    )

    kwargs.setdefault("verbose", True)

    importer = importer or objectio.ObjectImporter()
    for key, value in kwargs.iteritems():
        setattr(importer, key, value)

    importer.loads(json)

def copy_to_clipboard(*args, **kwargs):
    import clipboard
    json = export_json(*args, **kwargs)
    clipboard.copy(json.encode("utf-8"))

def paste_from_clipboard(**kwargs):
    import clipboard
    json = clipboard.paste().decode("utf-8")
    import_json(json, **kwargs)




#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			September 2008
"""
from cocktail.schema import Collection, Reference
from cocktail.translations import translations
from cocktail.html import Element, templates
from woost.views.uigeneration import backoffice_edit_control

Form = templates.get_class("cocktail.html.Form")


class ContentForm(Form):

    table_layout = False
    redundant_translation_labels = False
    base_ui_generators = [backoffice_edit_control]
    referer = None

    def create_member_display(self, obj, member, value, **context):
        context.setdefault("referer", self.persistent_object)
        return Form.create_member_display(self, obj, member, value, **context)

    def create_fieldset(self, group):
        fieldset = Form.create_fieldset(self, group)
        fieldset.set_client_param("group", group)
        return fieldset

    def get_group_label(self, group):

        label = Form.get_group_label(self, group)

        if not label and (not group or group == "default"):
            label = translations(self.schema.original_member.name)

        return label




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.iteration import first
from cocktail import schema
from woost.extensions.locations.location import Location


class CountryReference(schema.Reference):

    def __init__(self, *args, **kwargs):
        kwargs["type"] = Location
        kwargs.setdefault("default_order", "location_name")

        constraints = kwargs.get("relation_constraints")
        if constraints is None:
            constraints = {}
            kwargs["relation_constraints"] = constraints

        constraints["location_type"] = "country"

        schema.Reference.__init__(self, *args, **kwargs)

    def normalization(self, value):

        if isinstance(value, basestring):
            country = first(
                Location.select({
                    "location_type": "country",
                    "code": value.upper()
                })
            )
            if country is not None:
                value = country

        return value




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from cocktail.html import templates
from cocktail.html.uigeneration import (
    UIGenerator,
    EditControlGenerator,
    default_display,
    default_edit_control,
    default_search_control
)
from woost.models import Item

# Custom read only display for the backoffice
#------------------------------------------------------------------------------
def _collection_backoffice_display(
    ui_generator,
    obj,
    member,
    value,
    **context
):
    if member.is_persistent_relation:
        return "woost.views.ContentList"

backoffice_display = UIGenerator(
    "backoffice_display",
    base_ui_generators = [default_display],
    member_type_displays = {
        Item: "woost.views.ItemDisplay",
        schema.Member: "cocktail.html.TranslationDisplay",
        schema.Collection: _collection_backoffice_display,
        schema.BaseDateTime: "cocktail.html.RelativeTimeDisplay"
    }
)

backoffice_read_only_control = UIGenerator(
    "backoffice_read_only_control",
    base_ui_generators = [backoffice_display]
)

# Display for the 'Element' column of backoffice listings
#------------------------------------------------------------------------------
backoffice_element_column_display = UIGenerator(
    base_ui_generators = [backoffice_display],
    member_type_displays = {
        Item: "woost.views.ItemLabel"
    }
)

# Custom editable display for the backoffice
#------------------------------------------------------------------------------
def _collection_backoffice_edit_control(
    ui_generator,
    obj,
    member,
    value,
    **context
):
    if member.is_persistent_relation:
        return "woost.views.ItemCollectionEditor"
    else:
        return "cocktail.html.CollectionEditor"

def _reference_backoffice_edit_control(
    ui_generator,
    obj,
    member,
    value,
    **context
):
    if member.class_family is not None:
        display = templates.new("woost.views.Autocomplete")
        display.ajax_search = False
        display.show_types = False
        return display
    else:
        return "woost.views.ItemSelector"

backoffice_edit_control = EditControlGenerator(
    "backoffice_edit_control",
    base_ui_generators = [default_edit_control],
    member_type_displays = {
        schema.Collection: _collection_backoffice_edit_control,
        schema.Reference: _reference_backoffice_edit_control,
        schema.HTML: "woost.views.RichTextEditor"
    }
)

backoffice_edit_control.read_only_ui_generator = backoffice_read_only_control

# Custom selection controls for the item selector
#------------------------------------------------------------------------------
backoffice_selection_control = UIGenerator(
    "backoffice_selection_control",
    member_type_displays = {
        schema.Reference: "woost.views.Autocomplete"
    }
)

backoffice_integral_selection_control = UIGenerator(
    "backoffice_integral_selection_control",
    base_ui_generators = [backoffice_display]
)

# Custom search controls for the backoffice
#------------------------------------------------------------------------------
def _reference_backoffice_search_control(
    ui_generator,
    obj,
    member,
    value,
    **context
):
    if member.class_family is None and member.related_type is not None:
        display = templates.new("woost.views.Autocomplete")
        display.ajax_search_threshold = None
        member_schema = member.original_member.schema
        if not isinstance(member_schema, type) or not member_schema.__module__:
            display.ajax_url = (
                "/autocomplete/%s/QUERY"
                % member.related_type.get_qualified_name(include_ns = True)
            )
        return display

backoffice_search_control = EditControlGenerator(
    "backoffice_search_control",
    base_ui_generators = [
        default_search_control,
        backoffice_edit_control
    ],
    member_type_displays = {
        schema.Reference: _reference_backoffice_search_control
    }
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""

# Renderers
from woost.models.rendering.renderer import Renderer
from woost.models.rendering.chainrenderer import ChainRenderer
from woost.models.rendering.imagefilerenderer import ImageFileRenderer
from woost.models.rendering.imageurirenderer import ImageURIRenderer
from woost.models.rendering.videofilerenderer import VideoFileRenderer
from woost.models.rendering.pdfrenderer import PDFRenderer
from woost.models.rendering.htmlrenderer import HTMLRenderer
from woost.models.rendering.iconrenderer import IconRenderer

# Image effects
from woost.models.rendering.imagefactory import ImageFactory
from woost.models.rendering.imageeffects import (
    ImageEffect,
    ImageSize,
    Thumbnail,
    Crop,
    Fill,
    Rotate,
    Color,
    Brightness,
    Contrast,
    Sharpness,
    Frame,
    Shadow,
    ReducedOpacity,
    Fade,
    Watermark,
    Flip,
    Align
)

# Image requests and caching
from woost.models.rendering.request import (
    require_rendering,
    clear_image_cache,
    clear_image_cache_after_commit,
    BadRenderingRequest
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.modeling import cached_getter
from cocktail import schema
from cocktail.html.uigeneration import display_factory
from woost.controllers.backoffice.editstack import EditNode


class TranslationWorkflowRequestEditNode(EditNode):

    translated_values_key = "translated_values"

    @cached_getter
    def form_schema(self):
        form_schema = EditNode.form_schema(self)

        # Show the changelog for the request and its source item
        form_schema.add_member(
            schema.Member(
                "translation_workflow_changelog",
                editable = schema.READ_ONLY,
                member_group = "translation_request.changelog",
                backoffice_display = display_factory(
                    "woost.views.ChangeLog",
                    changes = [
                        change
                        for change in (
                              list(self.item.changes)
                            + list(self.item.translated_item.changes)
                        )
                        if change.is_explicit_change
                    ],
                    language_subset = (
                        self.item.source_language,
                        self.item.target_language
                    ),
                    exclude_member = lambda member:
                        member.name == "translation_requests",
                    show_all_changes = False
                )
            )
        )

        return form_schema

    @cached_getter
    def form_adapter(self):
        adapter = EditNode.form_adapter(self)

        editable = self.get_member_edit_mode(
            self.item.__class__.translated_values
        )
        if editable == schema.EDITABLE:
            adapter.export_rules.add_rule(
                TranslatedValuesExportRule(self.item),
                position = 0
            )
            adapter.import_rules.add_rule(
                TranslatedValuesImportRule(self.item),
                position = 0
            )
        return adapter


class TranslatedValuesRule(schema.Rule):

    key = "translated_values"

    def __init__(self, request):
        schema.Rule.__init__(self)
        self.request = request

    @cached_getter
    def translation_map(self):
        return dict(
            (member.name, self.key + "_" + member.name)
            for member in self.request.translated_item.__class__.iter_members()
            if member.translated and member.included_in_translation_workflow
        )


class TranslatedValuesExportRule(TranslatedValuesRule):

    @cached_getter
    def translated_values_adapter(self):
        adapter = schema.Adapter()
        adapter.implicit_copy = False
        adapter.copy(self.translation_map, properties = {
            "required": False,
            "translated": False,
            "unique": False,
            "member_group": "translation_request.translated_values"
        })
        return adapter

    def adapt_schema(self, context):
        if context.consume(self.key):

            # Preserve the 'original_member' and 'source_member' properties
            original_member = context.target_schema.original_member
            source_member = context.target_schema.source_member

            self.translated_values_adapter.export_schema(
                self.request.translated_item.__class__,
                context.target_schema
            )

            context.target_schema.original_member = original_member
            context.target_schema.source_member = source_member

    def adapt_object(self, context):
        if context.consume(self.key):
            item = self.request.translated_item
            values = self.request.translated_values
            for orig_key, trans_key in self.translation_map.iteritems():
                try:
                    value = values.get(orig_key)
                except AttributeError, KeyError:
                    value = item.get(orig_key, self.request.target_language)
                context.set(trans_key, value)


class TranslatedValuesImportRule(TranslatedValuesRule):

    def adapt_object(self, context):
        context.consume(self.key)
        values = self.request.translated_values
        if values is None:
            self.request.translated_values = values = {}
        for orig_key, trans_key in self.translation_map.iteritems():
            if context.consume(trans_key):
                values[orig_key] = context.get(trans_key)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.translations import translations
from cocktail import schema
from cocktail.persistence import PersistentList, PersistentMapping
from cocktail.html import templates
from woost.models import Item


class Survey(Item):

    members_order = [
        "title",
        "questions"
    ]

    title = schema.String(
        translated = True,
        required = True,
        indexed = True,
        unique = True,
        normalized_index = True,
        spellcheck = True,
        descriptive = True
    )

    questions = schema.Collection(
        items = "woost.extensions.surveys.survey.SurveyQuestion",
        bidirectional = True,
        integral = True,
        min = 1
    )

    def __init__(self, *args, **kwargs):
        Item.__init__(self, *args, **kwargs)
        self.__submissions = PersistentList()

    @property
    def submissions(self):
        return self.__submissions

    def add_submission(self, submission):
        self.__submissions.append(
            PersistentMapping(dict(submission))
        )

    def create_survey_schema(self):
        survey_schema = schema.Schema("SurveyForm")
        for question in self.questions:
            if question.active:
                question.init_survey_schema(survey_schema)
        return survey_schema


class SurveyQuestion(Item):

    instantiable = False
    visible_from_root = False
    results_view = None

    members_order = [
        "survey",
        "title",
        "active"
    ]

    survey = schema.Reference(
        type = "woost.extensions.surveys.survey.Survey",
        bidirectional = True
    )

    title = schema.String(
        translated = True,
        required = True,
        descriptive = True
    )

    active = schema.Boolean(
        required = True,
        default = True
    )

    def init_survey_schema(self, survey_schema):
        raise TypeError(
            "%s doesn't implement the init_survey_schema() method"
            % self
        )

    def create_survey_field(self, member_type, **member_kwargs):

        if "name" not in member_kwargs:
            member_kwargs["name"] = "question%d" % self.id

        if "translate_value" not in member_kwargs:
            member_kwargs["__translate__"] = (
                lambda language, **kwargs: translations(self, **kwargs)
            )

        return member_type(**member_kwargs)

    def create_results_view(self):
        if self.results_view:
            view = templates.new(self.results_view)
            view.question = self
            return view


class SurveyOptionsQuestion(SurveyQuestion):

    results_view = "woost.extensions.surveys.SurveyOptionsQuestionResultsView"

    options = schema.Collection(
        items = "woost.extensions.surveys.survey.SurveyOption",
        bidirectional = True,
        integral = True,
        min = 1
    )

    def init_survey_schema(self, survey_schema):
        survey_schema.add_member(
            self.create_survey_field(
                schema.Reference,
                type = SurveyOption,
                required = True,
                enumeration = [
                    option
                    for option in self.options
                    if option.active
                ],
                edit_control = "cocktail.html.RadioSelector"
            ),
            append = True
        )


class SurveyOption(Item):

    visible_from_root = False

    members_order = [
        "question",
        "title",
        "active"
    ]

    question = schema.Reference(
        type = "woost.extensions.surveys.survey.SurveyOptionsQuestion",
        bidirectional = True
    )

    title = schema.String(
        translated = True,
        required = True,
        descriptive = True
    )

    active = schema.Boolean(
        required = True,
        default = True
    )


class SurveyTextQuestion(SurveyQuestion):

    mandatory = schema.Boolean(
        required = True,
        default = True
    )

    def init_survey_schema(self, survey_schema):
        survey_schema.add_member(
            self.create_survey_field(
                schema.String,
                required = self.mandatory,
                edit_control = "cocktail.html.TextArea"
            ),
            append = True
        )




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.events import when
from cocktail.translations import iter_derived_languages
from cocktail import schema

def auto_enables_translations(cls):

    @when(cls.adding_translation)
    def handle_adding_translation(event):
        auto_enable_translations(
            event.language,
            event.source.translations,
            event.source.enabled_translations
        )

    @when(cls.removing_translation)
    def handle_removing_translation(event):
        auto_disable_translations(
            event.language,
            event.source.translations,
            event.source.enabled_translations
        )

    return cls

def auto_enable_translations(
    language,
    defined_translations,
    enabled_translations
):
    from woost.models import Configuration
    if language not in Configuration.instance.virtual_languages:
        schema.add(enabled_translations, language)

    for derived_language in iter_derived_languages(language):
        if derived_language not in defined_translations:
            auto_enable_translations(
                derived_language,
                defined_translations,
                enabled_translations
            )

def auto_disable_translations(
    language,
    defined_translations,
    enabled_translations
):
    try:
        schema.remove(enabled_translations, language)
    except (ValueError, KeyError):
        pass

    for derived_language in iter_derived_languages(language):
        if derived_language not in defined_translations:
            auto_disable_translations(
                derived_language,
                defined_translations,
                enabled_translations
            )




#-*- coding: utf-8 -*-
u"""
Site specific views.
"""
from pkg_resources import resource_filename
from cocktail.html import renderers, resource_repositories
renderers.default_renderer = renderers.html5_renderer

resource_repositories.define(
    "--SETUP-PACKAGE--",
    "/--SETUP-FLAT_WEBSITE_NAME--_resources",
    resource_filename("--SETUP-PACKAGE--.views", "resources")
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import os
from threading import local
from pkg_resources import resource_filename
from cocktail.modeling import GenericMethod
from cocktail.caching import Cache
from cocktail.controllers import context


class Application(object):

    __package = None
    __root = None
    __icon_resolver = None
    __custom_icon_repository = None

    __language = None
    __authentication = None

    installation_id = None
    installation_color = None

    splash = """\
Made with                     _
__        __ ___   ___   ___ | |_
\ \  __  / // _ \ / _ \ / __||  _|
 \ \/  \/ /| (_) | (_) |___ \| |__
  \__/\__/  \___/ \___/|____/\___/

http://woost.info
"""

    def __init__(self):
        self._thread_data = local()

    def path(self, *args):
        return os.path.join(self.root, *args)

    def _get_package(self):
        return self.__package

    def _set_package(self, package):
        self.__package = package
        self.__add_custom_icon_repository()

    package = property(_get_package, _set_package)

    def _get_root(self):
        if self.__root is None and self.__package:
            return resource_filename(self.__package, "")

        return self.__root

    def _set_root(self, root):
        self.__root = root
        self.__add_custom_icon_repository()

    root = property(_get_root, _set_root)

    def __add_custom_icon_repository(self):

        root = self.root
        package = self.package

        if root and package:
            repository = (
                self.path("views", "resources", "images", "icons"),
                "/" + self.package + "_resources/images/icons"
            )

            if (
                self.__custom_icon_repository is not None
                and self.__custom_icon_repository != repository
            ):
                self.icon_resolver.icon_repositories.remove(
                    self.__custom_icon_repository
                )

            self.__custom_icon_repository = repository
            self.icon_resolver.icon_repositories.insert(0, repository)

    @property
    def icon_resolver(self):
        if self.__icon_resolver is None:
            from woost.iconresolver import IconResolver
            self.__icon_resolver = IconResolver()
        return self.__icon_resolver

    # Language scheme
    def _get_language(self):
        if self.__language is None:
            from woost.languagescheme import LanguageScheme
            self.__language = LanguageScheme()
        return self.__language

    def _set_language(self, language):
        self.__language = language

    language = property(_get_language, _set_language)

    # Authentication scheme
    def _get_authentication(self):
        if self.__authentication is None:
            from woost.authenticationscheme import AuthenticationScheme
            self.__authentication = AuthenticationScheme()
        return self.__authentication

    def _set_authentication(self, authentication):
        self.__authentication = authentication

    authentication = property(_get_authentication, _set_authentication)

    # Caching
    cache = Cache()

    # URLs
    __url_resolver = None

    def _get_url_resolver(self):
        if self.__url_resolver is None:
            from woost.urlresolver import (
                URLResolver,
                HierarchicalURLScheme,
                DescriptiveIdURLScheme
            )
            url_resolver = URLResolver()
            url_resolver.add_url_scheme(HierarchicalURLScheme())
            url_resolver.add_url_scheme(DescriptiveIdURLScheme())
            self.__url_resolver = url_resolver
        return self.__url_resolver

    def _set_url_resolver(self, url_resolver):
        self.__url_resolver = url_resolver

    url_resolver = property(_get_url_resolver, _set_url_resolver)

    # Last error
    def _get_error(self):
        return getattr(self._thread_data, "error", None)

    def _set_error(self, error):
        self._thread_data.error = error

    error = property(_get_error, _set_error, doc =
        """Gets or sets the last error for the current context.

        "Context" is typically an HTTP request, but the property can also be
        used outside a web request/response cycle.

        .. type:: Exception
        """
    )

    traceback_link_style = "disabled"

    # Last traceback
    def _get_traceback(self):
        return getattr(self._thread_data, "traceback", None)

    def _set_traceback(self, traceback):
        self._thread_data.traceback = traceback

    traceback = property(_get_traceback, _set_traceback, doc =
        """Gets or sets the last traceback for the current context.

        "Context" is typically an HTTP request, but the property can also be
        used outside a web request/response cycle.

        .. type:: Traceback
        """
    )

    # Active user
    def _get_user(self):
        return getattr(self._thread_data, "user", None)

    def _set_user(self, user):
        self._thread_data.user = user

    user = property(_get_user, _set_user, doc =
        """Gets or sets the active user for the current context.

        "Context" is typically an HTTP request, but the property can also be
        used outside a web request/response cycle.

        .. type:: `woost.models.User`
        """
    )

    # Active website
    def _get_website(self):
        return getattr(self._thread_data, "website", None)

    def _set_website(self, website):
        self._thread_data.website = website

    website = property(_get_website, _set_website, doc =
        """Gets or sets the active website for the current context.

        "Context" is typically an HTTP request, but the property can also be
        used outside a web request/response cycle.

        .. type:: `woost.models.Website`
        """
    )

    # Active publishable
    def _get_publishable(self):
        return getattr(self._thread_data, "publishable", None)

    def _set_publishable(self, publishable):
        self._thread_data.publishable = publishable
        if self.original_publishable is None:
            self.navigation_point = get_navigation_point(publishable)

        # Required to preserve backward compatibility
        context["publishable"] = publishable

    publishable = property(_get_publishable, _set_publishable, doc =
        """Gets or sets the active publishable for the current context.

        "Context" is typically an HTTP request, but the property can also be
        used outside a web request/response cycle.

        .. type:: `woost.models.publishable`
        """
    )

    # Active original publishable
    def _get_original_publishable(self):
        return getattr(self._thread_data, "original_publishable", None)

    def _set_original_publishable(self, original_publishable):
        self._thread_data.original_publishable = original_publishable
        self.navigation_point = get_navigation_point(original_publishable)

        # Required to preserve backward compatibility
        context["original_publishable"] = original_publishable

    original_publishable = property(
        _get_original_publishable,
        _set_original_publishable,
        doc = """
        Gets or sets the active original_publishable for the current context.

        "Context" is typically an HTTP request, but the property can also be
        used outside a web request/response cycle.

        .. type:: `woost.models.original_publishable`
        """
    )

    # Active navigation point
    def _get_navigation_point(self):
        return getattr(self._thread_data, "navigation_point", None)

    def _set_navigation_point(self, navigation_point):
        self._thread_data.navigation_point = navigation_point

    navigation_point = property(_get_navigation_point, _set_navigation_point, doc =
        """Gets or sets the active navigation point for the current context.

        The navigation point is the publishable that should be highlighted as
        the active element in the navigation controls of the site's user
        interface. This will usually be the same as the active publishable,
        but not always: news items are a typical use case where the active
        publishable (the piece of news) and the highlighted page (the news
        listing) will differ.

        "Context" is typically an HTTP request, but the property can also be
        used outside a web request/response cycle.

        .. type:: `woost.models.navigation_point`
        """
    )


@GenericMethod
def get_navigation_point(self):
    return self

